import os
import warnings

import numpy as np
np.random.seed(0)
np.seterr(all='ignore')

from astropy.utils.exceptions import AstropyWarning

from .. import utils
from .. import filters
from .. import photoz
from .. import param

from . import test_filters
from . import test_templates

ez = None

# Single redshift for testing
z_spec = 1.0

# Additional catalog objects with random noise
NRND = 16

# S/N constant in F-lambda from F160W
# If None, then constant S/N across all bands
uncertainty_power = 2

def make_fake_catalog(SN=20):
    """
    Make a fake photometric catalog
    """
    
    data_path = test_filters.test_data_path()
    os.chdir(data_path)
    
    #### Generate data
    res = filters.FilterFile('filters/FILTER.RES.latest')
    templ = test_templates.read_template_ascii()
    
    ### WFC3 SED + K + IRAC
    f_numbers = [209, 211, 214, 217, 202, 203, 205, 269, 18, 19]
    f_list = [res[f] for f in f_numbers]
        
    ### Photometry from a single template
    fnu = templ.integrate_filter(f_list, z=z_spec)

    ### Norm to F160W
    i_f160 = -4
    flux_f160 = 1. # microJy
    fnu *= flux_f160/fnu[i_f160]
    
    ### Add noise
    #SN = 20
    efnu_f160w = (fnu/SN)[i_f160]
    lc = np.array([f.pivot for f in f_list])
    
    # Constant S/N in f-lambda
    if uncertainty_power is None:
        efnu = fnu/SN
    else:
        efnu = efnu_f160w*(lc/lc[i_f160])**uncertainty_power
    
    ### Make table
    tab = photoz.Table()
    tab['id'] = np.arange(NRND+1, dtype=int)+1
    tab['z_spec'] = z_spec
    tab['ra'] = 150.1
    tab['dec'] = 2.5
    
    ### Simpler filter names for catalog
    f_names = []
    for f in f_list:
        f_name = f.name.split(' ')[0].split('/')[-1].split('.dat')[0]
        f_name = f_name.replace('irac_tr','ch')
        f_name = f_name.replace('hawki_k','k').split('_')[0]
        f_names.append(f_name)
    
    ### Translate file
    translate_file = 'zphot.translate.test'
    
    with open(translate_file,'w') as fp:
        for i, f in enumerate(f_names):
            tab[f'f_{f}'] = fnu[i] + np.append(0, np.random.normal(size=NRND)*efnu[i])
            tab[f'e_{f}'] = efnu[i]

            fp.write(f'f_{f} F{f_numbers[i]}\n')
            fp.write(f'e_{f} E{f_numbers[i]}\n')
    
    tr = param.TranslateFile(translate_file)
    with open(translate_file + '.csv','w') as fp:
        fp.write(tr.to_csv())

    tr = param.TranslateFile(translate_file + '.csv')
    
    ### ASCII catalog
    cat_file = 'eazy_test.cat'
    tab.write(cat_file, overwrite=True, format='ascii.commented_header')
    tab.write(cat_file+'.fits', overwrite=True, format='fits')
        
    return tab, cat_file, translate_file


def test_full_photoz():
    """
    End-to-end test
    """
    global ez
        
    tab, cat_file, translate_file = make_fake_catalog(SN=20)
    
    data_path = test_filters.test_data_path()
    os.chdir(data_path)
    
    ### Parameters
    params = {}
    params['CATALOG_FILE'] = cat_file
    params['MAIN_OUTPUT_FILE'] = 'eazy_test'

    # Galactic extinction
    params['MW_EBV'] = 0.0

    params['Z_STEP'] = 0.01
    params['Z_MIN'] = z_spec - 0.5*(1+z_spec)
    params['Z_MAX'] = z_spec + 0.5*(1+z_spec)

    params['SYS_ERR'] = 0.02

    params['PRIOR_ABZP'] = 23.9 # uJy 
    params['PRIOR_FILTER'] = 205 # f160W
    params['PRIOR_FILE'] = 'templates/prior_F160W_TAO.dat'

    params['FILTERS_RES'] = 'filters/FILTER.RES.latest'
    params['TEMPLATES_FILE'] = 'templates/fsps_full/fsps_QSF_12_v3.param'
    params['VERBOSITY'] = 1
    params['FIX_ZSPEC'] = False
    
    ### Initialize object
    ez = photoz.PhotoZ(param_file=None, translate_file=translate_file, 
                         zeropoint_file=None, params=params, load_prior=True, 
                         load_products=False)
    
    # FITS catalog
    params['CATALOG_FILE'] = cat_file+'.fits'
    ez = photoz.PhotoZ(param_file=None, translate_file=translate_file, 
                         zeropoint_file=None, params=params, load_prior=True, 
                         load_products=False)


def test_photoz_methods():
    """
    Test methods on `~eazy.photoz.PhotoZ` object.
    """
    global ez
    
    ### Catalog subset
    ez.fit_catalog(idx=np.where(ez.cat['id'] < 2)[0], fitter='nnls')
    
    ### Full catalog, fitting methods
    ez.fit_catalog(fitter='lstsq')
    ez.fit_catalog(fitter='bounded')
    
    # Serial
    ez.fit_catalog(fitter='nnls', n_proc=0)
    chi2_serial = ez.chi2_fit
    coeffs_serial = ez.fit_coeffs
    
    # Parallel
    ez.fit_catalog(fitter='nnls', n_proc=4)
    assert(np.allclose(ez.chi2_fit, chi2_serial))
    assert(np.allclose(ez.fit_coeffs, coeffs_serial))
      
    ###### Methods
    
    # Specified zbest
    ez.fit_at_zbest(zbest=np.full(NRND+1, z_spec), 
              prior=False, beta_prior=False, 
              get_err=False, clip_wavelength=1100, fitter='nnls', 
              selection=None, n_proc=0, par_skip=10000)
    
    # default zbest
    ez.fit_at_zbest(zbest=None, prior=False, beta_prior=False, 
            get_err=False, clip_wavelength=1100, fitter='nnls', 
            selection=None, n_proc=0, par_skip=10000)
    
    # priors
    for prior in [True, False]:
        for beta_prior in [True, False]:
            ez.fit_at_zbest(zbest=None, prior=prior, beta_prior=beta_prior, 
                      get_err=False, clip_wavelength=1100, fitter='nnls', 
                      selection=None, n_proc=0, par_skip=10000)
    
    # Peak-finder
    peaks, numpeaks = ez.find_peaks()
    assert(np.allclose(numpeaks, 1))
    assert(np.allclose(ez.zgrid[peaks[0][0]], z_spec, atol=0.01*(1+z_spec)))
    
    return ez


def test_sps_parameters():
    """
    Derived parameters
    """
    global ez
    
    import astropy.units as u
    
    ### Run all photo-zs
    ez.fit_catalog(fitter='nnls')
        
    ### SPS parameters
    
    # Parameters normalized by V band
    sps = ez.sps_parameters(template_fnu_units=None, simple=True)
    
    fnu = (1*u.solLum / u.Hz)
    sps = ez.sps_parameters(template_fnu_units=fnu, simple=True)
    
    # Full RF-colors with filter weighting
    zout, hdu = ez.standard_output(zbest=None, rf_pad_width=0.5, rf_max_err=2, 
                                   prior=True, beta_prior=True, simple=False,
                                   save_fits=False)

    # "Simple" best-fit template RF colors
    zout, hdu = ez.standard_output(zbest=None, rf_pad_width=0.5, rf_max_err=2, 
                                   prior=True, beta_prior=True, simple=True, 
                                   save_fits=False)
    
    assert(np.allclose(zout['z_phot'][0], z_spec, atol=0.1*(1+z_spec)))

    coeffs_norm = ez.coeffs_best[0,:]/ez.coeffs_best[0,:].max()     
    assert(np.argmax(coeffs_norm) == 0)
    assert(np.sum(coeffs_norm) < 1.1)
    
    # All zout data
    # zdict = {}
    # for k in zout.colnames:
    #     zdict[k] = zout[k][0]
    
    zdict = {'nusefilt': 10,
             'z_ml': 0.99616235,
             'z_ml_chi2': 0.013447836,
             'z_ml_risk': 0.0105553605,
             'lc_min': 3353.6304006459895,
             'lc_max': 45020.33785230743,
             'z_phot': 0.99616235,
             'z_phot_chi2': 0.013447836,
             'z_phot_risk': 0.0105553605,
             'z_min_risk': 0.9937155,
             'min_risk': 0.010250151,
             'z_raw_chi2': 0.9937155,
             # 'raw_chi2': 0.035614725,
             'z025': 0.92501247,
             'z160': 0.9604295,
             'z500': 0.99208033,
             'z840': 1.0187114,
             'z975': 1.0420052,
             'restU': 0.41460526,
             'restU_err': 0.01217702,
             'restB': 0.8223915,
             'restB_err': 0.027577162,
             'restV': 0.92202765,
             'restV_err': 0.017819434,
             'restJ': 1.024555,
             'restJ_err': 0.05461645,
             'dL': 6575.8372348364455,
             'mass': 1338132577.7487125,
             'sfr': 0.026515421690098212,
             'Lv': 3418389791.239653,
             'LIR': 438179193.31513166,
             'MLv': 0.39145113912344354,
             'Av': 0.06295947926487588,
             # 'rest270': 0.11133574,
             # 'rest270_err': 0.007641867,
             # 'rest274': 0.23238972,
             # 'rest274_err': 0.008679345,
             # 'rest120': 0.12516989,
             # 'rest120_err': 0.005393833,
             # 'rest121': 0.1816069,
             # 'rest121_err': 0.00364957,
             # 'rest156': 0.3724664,
             # 'rest156_err': 0.014633045,
             # 'rest157': 0.86651146,
             # 'rest157_err': 0.018754214,
             # 'rest158': 0.94490474,
             # 'rest158_err': 0.027536243,
             # 'rest159': 0.997915,
             # 'rest159_err': 0.023829281,
             # 'rest160': 1.0238949,
             # 'rest160_err': 0.0475851,
             # 'rest161': 1.024555,
             # 'rest161_err': 0.05461645,
             # 'rest162': 1.010895,
             # 'rest162_err': 0.06887752,
             # 'rest163': 0.7563232,
             # 'rest163_err': 0.06583378,
             'DISTMOD': 43.33921454218198,
         }
    
    for k in zdict:
        if '_err' in k:
            assert(np.allclose(zout[k][0], zdict[k], rtol=1.0))
        else:
            assert(np.allclose(zout[k][0], zdict[k], rtol=1.0))

    # confirm that zout['z_phot'] == zout['z_ml']
    assert(np.all(zout['z_ml'] == zout['z_phot']))
        
    ### user-specified zbest
    zuser = np.full(NRND+1, z_spec)
    z2, _ = ez.standard_output(zbest=zuser, rf_pad_width=0.5, rf_max_err=2, 
                               prior=True, beta_prior=True, simple=True,
                               save_fits=False)
    
    # confirm that z2 has 'z_ml' and 'z_phot' columns and they're different 
    assert(np.all(z2['z_ml'] != z2['z_phot']))

    # confirm that z2['z_ml'] == zout['z_phot']
    assert(np.all(z2['z_ml'] == zout['z_phot']))
    
    # zphot is now the user-specified redshift
    assert(np.allclose(z2['z_phot'], zuser, rtol=1.e-2))
    
    # SPS parameters are different, as calculated for zuser
    assert(np.all(z2['mass'] != zout['mass']))
    assert(np.all(z2['sfr'] != zout['sfr']))


def test_load_products():
    """
    Save and read products
    """
    global ez
    
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyWarning)
    
        zout, hdu = ez.standard_output(zbest=None, rf_pad_width=0.5,
                                       rf_max_err=2,
                                       prior=True, beta_prior=True,
                                       simple=True,
                                       save_fits=2)
                                   
    new = photoz.PhotoZ(param_file='eazy_test.zphot.param',
                        translate_file='eazy_test.zphot.translate', 
                        zeropoint_file='eazy_test.zphot.zeropoint', 
                        load_prior=True, 
                        load_products=False)
    
    assert(len(new.param.params) == len(ez.param.params))
    
    bool_param = utils.bool_param
    for k in ez.param.params:
        assert(bool_param(ez.param.params[k]) == 
               bool_param(new.param.params[k]))

    new.load_products()
    assert(new.ZML_WITH_PRIOR == ez.ZML_WITH_PRIOR)
    assert(new.ZML_WITH_BETA_PRIOR == ez.ZML_WITH_BETA_PRIOR)
    
    has_chi2_init = (ez.chi2_fit != 0).sum(axis=1) > 0 
    has_chi2 = (new.chi2_fit != 0).sum(axis=1) > 0 
    assert(has_chi2_init.sum() == has_chi2.sum())

    assert(np.allclose(new.coeffs_best, ez.coeffs_best))

    assert(np.allclose(new.fit_coeffs, ez.fit_coeffs))

    assert(np.allclose(new.prior_data, ez.prior_data))

    assert(np.allclose(ez.lnp[has_chi2,:], new.lnp[has_chi2,:]))


def test_fit_stars():
    """
    Fit phoenix star library for Star/Galaxy separation
    """
    global ez
    res = ez.fit_phoenix_stars()
    assert (np.allclose(ez.star_chi2[0,0], 3191.3662))
    
    sub_index = np.array([0,1])
    sub_bool = np.zeros(ez.NOBJ, dtype=bool)
    sub_bool[:2] = True
    
    for subset in [sub_index, sub_bool]:
        res = ez.fit_phoenix_stars(subset=subset)
        assert 'subset' in res
        assert res['star_min_ix'].shape == (2,)
        assert res['star_tnorm'].shape == (2, ez.NSTAR)
        assert (np.allclose(res['star_chi2'][0,0], 3191.3662))


def test_photoz_figures():
    """
    Figures generated with PhotoZ object
    """
    import matplotlib.pyplot as plt
    
    global ez
    
    ### SED figure
    fig, data = ez.show_fit(id=0, id_is_idx=True, show_fnu=False)
    fig.savefig('eazy_test.sed.png', dpi=72)
    
    assert(isinstance(fig, plt.Figure))
    assert(isinstance(data, dict))
    
    fig = ez.show_fit(id=1, show_fnu=False)
    fig = ez.show_fit(id=1, show_fnu=True)
    fig = ez.show_fit(id=1, show_fnu=2)

    fig = ez.show_fit(id=1, show_fnu=False, zshow=z_spec)

    fig = ez.show_fit(id=1, show_components=True)
    
    fig = ez.zphot_zspec()
    fig = ez.zphot_zspec(zmin=0, zmax=2)
    fig.savefig('eazy_test.zphot_zspec.png', dpi=72)
    
    plt.close('all')


def test_zeropoint_residuals():
    """
    """
    global ez

    ez.fit_catalog(fitter='nnls')
    res = ez.residuals()


def test_hdf5():
    """
    Test HDF5 save / recover state
    """
    import matplotlib.pyplot as plt
    plt.ioff()

    global ez
    
    from .. import hdf5
    try:
        import h5py
    except ImportError:
        return None
    
    hdf5.write_hdf5(ez, h5file='test.hdf5', include_fit_coeffs=False, 
                    include_templates=True)
    
    new_ez = hdf5.initialize_from_hdf5('test.hdf5')
    
    assert(np.allclose(ez.fnu, new_ez.fnu))
    assert(np.allclose(ez.efnu, new_ez.efnu))
    assert(np.allclose(ez.prior_mags, new_ez.prior_mags))
    assert(np.allclose(ez.zml, new_ez.zml))
    assert(np.allclose(ez.zbest, new_ez.zbest))
    assert(np.allclose(ez.lnp, new_ez.lnp, rtol=1.e-4))

    _  = new_ez.show_fit(10)
    plt.close('all')

    # Compact viewer
    h5 = hdf5.Viewer('test.hdf5')

    assert(h5.NOBJ == new_ez.NOBJ)
    assert(h5.NTEMP == new_ez.NTEMP)
    assert(h5.NFILT == new_ez.NFILT)
    assert(h5.NZ == new_ez.NZ)
    assert(np.allclose(h5.zp, new_ez.zp))

    _cat = h5.get_catalog()
    _ = h5.show_fit(10)
    plt.close('all')


def test_cleanup():
    
    if os.path.exists('filters/FILTER.RES.latest.npy'):
        os.remove('filters/FILTER.RES.latest.npy')
        


import numpy as np

from .. import igm

def test_igm():
    """
    Test IGM module (Inoue14)
    """
    
    igm_obj = igm.Inoue14()
    
    # Test against result from a particular version
    # (db97f839cf8afe4a22c31c5d6195fd707ba4de32)
    z = 3.0
    rest_wave = np.arange(850, 1251, 50)
    igm_val = np.array([0.33537573, 0.54634578, 0.74207249, 0.74194787, 
                        0.79182545, 0.75792504, 0.72135181, 0.68233589, 1.0])
    
    assert(np.allclose(igm_obj.full_IGM(z, rest_wave*(1+z)), igm_val, rtol=1.e-2))
    
    # With scaling               
    scale_tau = 2.    
    igm_obj.scale_tau = scale_tau
    
    igm_scaled = np.exp(2*np.log(igm_val))
    assert(np.allclose(igm_obj.full_IGM(z, rest_wave*(1+z)), igm_scaled, 
                       rtol=1.e-2))
    
    igm_obj = igm.Inoue14(scale_tau=scale_tau)
    assert(np.allclose(igm_obj.full_IGM(z, rest_wave*(1+z)), igm_scaled, 
                       rtol=1.e-2))
    
    


"""
Tools for making FSPS templates
"""
import os
from collections import OrderedDict

import numpy as np
import astropy.units as u
from astropy.cosmology import WMAP9

FLAM_CGS = u.erg/u.second/u.cm**2/u.Angstrom
LINE_CGS = 1.e-17*u.erg/u.second/u.cm**2

try:
    from dust_attenuation.baseclasses import BaseAttAvModel
except:
    BaseAttAvModel = object

from astropy.modeling import Parameter
import astropy.units as u

try:
    from fsps import StellarPopulation
except:
    # Broken, but imports
    StellarPopulation = object

from . import utils
from . import templates

DEFAULT_LABEL = 'fsps_tau{tau:3.1f}_logz{logzsol:4.2f}_tage{tage:4.2f}_av{Av:4.2f}'

WG00_DEFAULTS = dict(geometry='shell', dust_type='mw', 
                   dust_distribution='homogeneous')

__all__ = ["Zafar15", "ExtinctionModel", "SMC", "Reddy15", "KC13",
           "ParameterizedWG00", "ExtendedFsps", "fsps_line_info", 
           "wuyts_line_Av"]


class ArrayExtCurve(BaseAttAvModel):
    """
    Alam interpolated from arrays
    """
    name = 'Array'
    #bump_ampl = 1.
        
    Rv = 2.21 # err 0.22
    
    xarray = np.arange(0.09, 2.2, 0.01)
    yarray = xarray*0.+1
    left=None
    right=None
    
    def Alam(self, mu):
        """
        klam, eq. 1
        """
        Alam = np.interp(mu, self.xarray, self.yarray, 
                         left=self.left, right=self.right)
        return Alam
            
    def evaluate(self, x, Av):
       
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.micron
        else:
            xin = np.atleast_1d(x)
        
        mu = xin.to(u.micron).value

        alam = self.Alam(mu) #*self.Rv
        return np.maximum(alam*Av, 0.)


class Zafar15(BaseAttAvModel):
    """
    Quasar extinction curve from Zafar et al. (2015)
        
    https://ui.adsabs.harvard.edu/abs/2015A%26A...584A.100Z/abstract
    """
    name = 'Zafar+15'
    #bump_ampl = 1.
        
    Rv = 2.21 # err 0.22
    
    @staticmethod
    def Alam(mu, Rv):
        """
        klam, eq. 1
        """
        x = 1/mu
                
        # My fit
        coeffs = np.array([0.05694421, 0.57778243, -0.12417444])
        Alam = np.polyval(coeffs, x)*2.21/Rv
        
        # Only above x > 5.90
        fuv = x > 5.90
        if fuv.sum() > 0:
            Afuv = 1/Rv*(-4.678+2.355*x + 0.622*(x-5.90)**2) + 1.
            Alam[fuv] = Afuv[fuv]
            
        return Alam
            
    def evaluate(self, x, Av):
       
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.micron
        else:
            xin = np.atleast_1d(x)
        
        mu = xin.to(u.micron).value

        alam = self.Alam(mu, self.Rv) #*self.Rv

        # Rv = Av/EBV
        # EBV=Av/Rv
        # Ax = Alam/Av
        # 
        # klam = Alam/EBV
        # Alam = klam*EBV = klam*Av/Rv
        return np.maximum(alam*Av, 0.)

class ExtinctionModel(BaseAttAvModel):
    """
    Modify `dust_extinction.averages.G03_SMCBar` to work as Att
    """     
    #from dust_extinction.averages import G03_SMCBar
    #SMCBar = G03_SMCBar()
    curve_type = 'smc'
    init_curve = None
    
    #@property
    def _curve_model(self):
        
        if self.init_curve == self.curve_type:
            return 0
            
        if self.curve_type.upper() == 'SMC':
            from dust_extinction.averages import G03_SMCBar as curve
        elif self.curve_type.upper() == 'LMC':
            from dust_extinction.averages import G03_LMCAvg as curve
        elif self.curve_type.upper() in ['MW','F99']:
            from dust_extinction.parameter_averages import F99 as curve
        else:
            raise ValueError(f'curve_type {self.curve_type} not recognized')
            
        self.curve = curve()
        self.init_curve = self.curve_type
        
    def evaluate(self, x, Av):
        
        self._curve_model()
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.Angstrom
        else:
            xin = np.atleast_1d(x)
        
        xinv = 1./xin.to(u.micron)

        curve = self.curve
        xr = [x for x in curve.x_range]
        xr[0] *= 1.001
        xr[1] *= 0.999
        print('xxx', xr)
        
        if 'Rv' in curve.param_names:
            klam = curve.evaluate(1/np.clip(xinv, 
                                    xr[0]/u.micron, xr[1]/u.micron), 
                                    Rv=curve.Rv)
        else:
            klam = curve.evaluate(1/np.clip(xinv, 
                                    xr[0]/u.micron, xr[1]/u.micron))
            
        return klam*Av

class SMC(BaseAttAvModel):
    """
    Modify `dust_extinction.averages.G03_SMCBar` to work as Att
    """     
    from dust_extinction.averages import G03_SMCBar
    SMCBar = G03_SMCBar()
    
    bump_ampl = Parameter(description="Amplitude of UV bump",
                      default=0., min=0., max=10.)
    
    bump_gamma = 0.04
    bump_x0 = 0.2175
    
    
    def uv_bump(self, mu, bump_ampl):
        """
        Drude profile for computing the UV bump.

        Parameters
        ----------
        x: np array (float)
           expects wavelengths in [micron]

        x0: float
           Central wavelength of the UV bump (in microns).

        gamma: float
           Width (FWHM) of the UV bump (in microns).

        ampl: float
           Amplitude of the UV bump.

        Returns
        -------
        np array (float)
           lorentzian-like Drude profile

        Raises
        ------
        ValueError
           Input x values outside of defined range

        """
        return bump_ampl * (mu**2 * self.bump_gamma**2 /
                       ((mu**2 - self.bump_x0**2)**2 + 
                         mu**2 * self.bump_gamma**2))
    
    
    def evaluate(self, x, Av, bump_ampl):

        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.Angstrom
        else:
            xin = np.atleast_1d(x)

        xinv = 1./xin.to(u.micron)

        klam = self.SMCBar.evaluate(1/np.clip(xinv, 
                                    0.301/u.micron, 9.99/u.micron))
        
        if bump_ampl > 0:
            klam += self.uv_bump(xin.to(u.micron).value, bump_ampl)
            
        return klam*Av
                

class Reddy15(BaseAttAvModel):
    """
    Attenuation curve from Reddy et al. (2015)
    
    With optional UV bump
    
    https://ui.adsabs.harvard.edu/abs/2015ApJ...806..259R/abstract
    """
    name = 'Reddy+15'
    #bump_ampl = 1.
    
    bump_ampl = Parameter(description="Amplitude of UV bump",
                      default=2., min=0., max=10.)
    
    bump_gamma = 0.04
    bump_x0 = 0.2175
    
    Rv = 2.505
    
    @staticmethod
    def _left(mu):
        """
        klam, mu < 0.6 micron
        """
        return -5.726 + 4.004/mu - 0.525/mu**2 + 0.029/mu**3 + 2.505
    
    
    @staticmethod
    def _right(mu):
        """
        klam, mu > 0.6 micron
        """
        return -2.672 - 0.010/mu + 1.532/mu**2 - 0.412/mu**3 + 2.505
    
    @property
    def koffset(self):
        """
        Force smooth transition at 0.6 micron
        """
        return self._left(0.6) - self._right(0.6)
        
    def evaluate(self, x, Av, bump_ampl):
       
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.Angstrom
        else:
            xin = np.atleast_1d(x)
        
        mu = xin.to(u.micron).value
        left =  mu < 0.6
        klam = mu*0.
        # Reddy Eq. 8
        kleft = self._left(mu)
        kright = self._right(mu)
        
        klam[left] = self._left(mu[left])
        klam[~left] = self._right(mu[~left]) + self.koffset

        # Rv = Av/EBV
        # EBV=Av/Rv
        # klam = Alam/EBV
        # Alam = klam*EBV = klam*Av/Rv
        return np.maximum((klam + self.uv_bump(mu, bump_ampl))*Av/self.Rv, 0.)
    
    def uv_bump(self, mu, bump_ampl):
        """
        Drude profile for computing the UV bump.

        Parameters
        ----------
        x: np array (float)
           expects wavelengths in [micron]

        x0: float
           Central wavelength of the UV bump (in microns).

        gamma: float
           Width (FWHM) of the UV bump (in microns).

        ampl: float
           Amplitude of the UV bump.

        Returns
        -------
        np array (float)
           lorentzian-like Drude profile

        Raises
        ------
        ValueError
           Input x values outside of defined range

        """
        return bump_ampl * (mu**2 * self.bump_gamma**2 /
                       ((mu**2 - self.bump_x0**2)**2 + 
                         mu**2 * self.bump_gamma**2))

class KC13(BaseAttAvModel):
    """
    Kriek & Conroy (2013) attenuation model, extends Noll 2009 with UV bump 
    amplitude correlated with the slope, delta.
    
    Slightly different from KC13 since the N09 model uses Leitherer (2002) 
    below 1500 Angstroms.
    
    """
    name = 'Kriek+Conroy2013'
    
    delta = Parameter(description="delta: slope of the power law",
                      default=0., min=-3., max=3.)
    
    #extra_bump = 1.
    extra_params = {'extra_bump':1., 'beta':-3.2, 'extra_uv':-0.4}
    
    # Big range for use with FSPS
    x_range = [0.9e-4, 2.e8]
    
    def _init_N09(self):
        from dust_attenuation import averages, shapes, radiative_transfer

        # Allow extrapolation
        #shapes.x_range_N09 = [0.9e-4, 2.e8] 
        #averages.x_range_C00 = [0.9e-4, 2.e8]
        #averages.x_range_L02 = [0.9e-4, 0.18]
        shapes.C00.x_range = self.x_range
        shapes.N09.x_range = self.x_range 
        if self.x_range[0] < 0.18:
            shapes.L02.x_range = [self.x_range[0], 0.18] 
        else:
            shapes.L02.x_range = [0.097, 0.18] 
                    
        self.N09 = shapes.N09()
                
    def evaluate(self, x, Av, delta):
        import dust_attenuation
        
        if not hasattr(self, 'N09'):
            self._init_N09()
                    
        #Av = np.polyval(self.coeffs['Av'], tau_V)
        x0 = 0.2175
        gamma = 0.0350
        ampl = (0.85 - 1.9*delta)*self.extra_params['extra_bump']
        
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.Angstrom
        else:
            xin = x
        
        wred = np.array([2.199e4])*u.Angstrom
        
        if self.N09.param_names[0] == 'x0':                
            Alam = self.N09.evaluate(xin, x0, gamma, ampl, delta, Av)
            Ared = self.N09.evaluate(wred, x0, gamma, ampl, delta, Av)[0]
        else:
            Alam = self.N09.evaluate(xin, Av, x0, gamma, ampl, delta)
            Ared = self.N09.evaluate(wred, Av, x0, gamma, ampl, delta)[0]
            
        
        # Extrapolate with beta slope
        red = xin > wred[0]
        if red.sum() > 0:
            Alam[red] = Ared*(xin[red]/wred[0])**self.extra_params['beta']
        
        blue = xin < 1500*u.Angstrom
        if blue.sum() > 0:
            plblue = np.ones(len(xin))
            wb = xin[blue].to(u.Angstrom).value/1500
            plblue[blue] = wb**self.extra_params['extra_uv']
            Alam *= plblue
            
        return Alam
        
class ParameterizedWG00(BaseAttAvModel):
    
    coeffs = {'Av': np.array([-0.001,  0.026,  0.643, -0.016]),
              'x0': np.array([ 3.067e-19, -7.401e-18,  6.421e-17, -2.370e-16,  
                               3.132e-16, 2.175e-01]),
              'gamma': np.array([ 2.101e-06, -4.135e-05,  2.719e-04, 
                                 -7.178e-04, 3.376e-04, 4.270e-02]),
              'ampl': np.array([-1.906e-03,  4.374e-02, -3.501e-01, 
                                 1.228e+00, -2.151e+00, 8.880e+00]),
              'slope': np.array([-4.084e-05,  9.984e-04, -8.893e-03,  
                                 3.670e-02, -7.325e-02, 5.891e-02])}
    
    # Turn off bump
    include_bump = 0.25
    
    wg00_coeffs = {'geometry': 'shell', 
                   'dust_type': 'mw',
                   'dust_distribution': 'homogeneous'}
    
    name = 'ParameterizedWG00'
        
    # def __init__(self, Av=1.0, **kwargs):
    #     """
    #     Version of the N09 curves fit to the WG00 curves up to tauV=10
    #     """        
    #     from dust_attenuation import averages, shapes, radiative_transfer
    #     
    #     # Allow extrapolation
    #     shapes.x_range_N09 = [0.01, 1000] 
    #     averages.x_range_C00 = [0.01, 1000]
    #     averages.x_range_L02 = [0.01, 0.18]
    #     
    #     self.N09 = shapes.N09()
    
    def _init_N09(self):
        from dust_attenuation import averages, shapes, radiative_transfer

        # Allow extrapolation
        shapes.x_range_N09 = [0.009, 2.e8] 
        averages.x_range_C00 = [0.009, 2.e8]
        averages.x_range_L02 = [0.009, 0.18]

        self.N09 = shapes.N09()
        
    def get_tau(self, Av):
        """
        Get the WG00 tau_V for a given Av
        """
        tau_grid = np.arange(0, 10, 0.01)
        av_grid = np.polyval(self.coeffs['Av'], tau_grid)
        return np.interp(Av, av_grid, tau_grid, left=0., right=tau_grid[-1])
        
    def evaluate(self, x, Av):
        
        import dust_attenuation
        
        if not hasattr(self, 'N09'):
            self._init_N09()
            
        tau_V = self.get_tau(Av)
        
        #Av = np.polyval(self.coeffs['Av'], tau_V)
        x0 = np.polyval(self.coeffs['x0'], tau_V)
        gamma = np.polyval(self.coeffs['gamma'], tau_V)
        if self.include_bump:
            ampl = np.polyval(self.coeffs['ampl'], tau_V)*self.include_bump
        else:
            ampl = 0.
            
        slope = np.polyval(self.coeffs['slope'], tau_V)
        
        if not hasattr(x, 'unit'):
            xin = np.atleast_1d(x)*u.Angstrom
        else:
            xin = x
        
        if self.N09.param_names[0] == 'x0':                
            return self.N09.evaluate(xin, x0, gamma, ampl, slope, Av)
        else:
            return self.N09.evaluate(xin, Av, x0, gamma, ampl, slope)


def fsps_line_info(wlimits=None):
    """
    Read FSPS line list
    """
    try:
        info_file = os.path.join(os.getenv('SPS_HOME'), 'data/emlines_info.dat')
        with open(info_file, 'r') as f:
            lines = f.readlines()
    except:
        return [], []
        
    waves = np.array([float(l.split(',')[0]) for l in lines])
    names = np.array([l.strip().split(',')[1].replace(' ','') for l in lines])
    if wlimits is not None:
        clip = (waves > wlimits[0]) & (waves < wlimits[1])
        waves = waves[clip]
        names = names[clip]
        
    return waves, names


DEFAULT_LINES = fsps_line_info(wlimits=[1200, 1.9e4])[0]

BOUNDS = {}
BOUNDS['tage'] = [0.03, 12, 0.05]
BOUNDS['tau'] = [0.03, 2, 0.05]
BOUNDS['zred'] = [0.0, 13, 1.e-4]
BOUNDS['Av'] = [0.0, 15, 0.05]
BOUNDS['gas_logu'] = [-4, 0, 0.05]
BOUNDS['gas_logz'] = [-2, 0.3, 0.05]
BOUNDS['logzsol'] = [-2, 0.3, 0.05]
BOUNDS['sigma_smooth'] = [100, 500, 0.05]


def wuyts_line_Av(Acont):
    """
    Wuyts prescription for extra extinction towards nebular emission
    """
    return Acont + 0.9*Acont - 0.15*Acont**2


class ExtendedFsps(StellarPopulation):
    """
    Extended functionality for the `fsps.StellarPopulation` object
    """
    
    lognorm_center = 0.
    lognorm_logwidth = 0.05
    is_lognorm_sfh = False
    lognorm_fburst = -30
    
    cosmology = WMAP9
    scale_lyman_series = 0.1
    scale_lines = OrderedDict()
    
    line_av_func = None
    
    # Smoothing parameters for fitting
    lsf_func = None
    FFT_SMOOTH = False
    
    #_meta_bands = ['v']
    
    @property
    def izmet(self):
        """
        Get zmet index for nearest ``self.zlegend`` value to ``loggzsol``.
        """
        NZ = len(self.zlegend)
        logzsol = self.params['logzsol']
        zi = np.interp(logzsol, np.log10(self.zlegend/0.019), np.arange(NZ))
        return np.clip(np.asarray(np.round(zi),dtype=int), 0, NZ-1)
    
    @property
    def fsps_ages(self):
        """
        (linear) ages of the FSPS SSP age grid, Gyr
        """
        if hasattr(self, '_fsps_ages'):
            return self._fsps_ages
        
        _ = self.get_spectrum()  
        fsps_ages = 10**(self.log_age-9)
        self._fsps_ages = fsps_ages
        return fsps_ages
        
    def set_lognormal_sfh(self, min_sigma=3, verbose=False, **kwargs):
        """
        Set lognormal tabular SFH
        """
        try:
            from grizli.utils_c.interp import interp_conserve_c as interp_func
        except:
            interp_func = utils.interp_conserve
        
        if 'lognorm_center' in kwargs:
            self.lognorm_center = kwargs['lognorm_center']
        
        if 'lognorm_logwidth' in kwargs:
            self.lognorm_logwidth = kwargs['lognorm_logwidth']
                
        if self.is_lognorm_sfh:
            self.params['sfh'] = 3
        
        if verbose:
            msg = 'lognormal SFH ({0}, {1}) [sfh3={2}]'
            print(msg.format(self.lognorm_center, self.lognorm_logwidth, 
                             self.is_lognorm_sfh))
                
        xages = np.logspace(np.log10(self.fsps_ages[0]), 
                            np.log10(self.fsps_ages[-1]), 2048)

        mu = self.lognorm_center#*np.log(10)
        # sfh = 1./t*exp(-(log(t)-mu)**2/2/sig**2)
        logn_sfh = 10**(-(np.log10(xages)-mu)**2/2/self.lognorm_logwidth**2)
        logn_sfh *= 1./xages 
        
        # Normalize
        logn_sfh *= 1.e-9/(self.lognorm_logwidth*np.sqrt(2*np.pi*np.log(10)))
        
        self.set_tabular_sfh(xages, logn_sfh) 
        self._lognorm_sfh = (xages, logn_sfh)
    
    def lognormal_integral(self, tage=0.1, **kwargs):
        """
        Integral of lognormal SFH up to t=tage
        """
        from scipy.special import erfc
        mu = self.lognorm_center*np.log(10)
        sig = self.lognorm_logwidth*np.sqrt(np.log(10))
        cdf = 0.5*erfc(-(np.log(tage)-mu)/sig/np.sqrt(2)) 
        return cdf
        
    def _set_extend_attrs(self, line_sigma=50, lya_sigma=200, **kwargs):
        """
        Set attributes on `fsps.StellarPopulation` object used by `narrow_lines`.

        sigma : line width (FWHM/2.35), km/s.
        lya_sigma : width for Lyman-alpha

        Sets `emline_dlam`, `emline_sigma` attributes.

        """
        
        # Line widths, native FSPS and new 
        wave, line = self.get_spectrum(tage=1., peraa=True)
        dlam = np.diff(wave)
        self.emline_dlam = [np.interp(w, wave[1:], dlam) 
                            for w in self.emline_wavelengths] # Angstrom
        self.emline_sigma = [line_sigma for w in self.emline_wavelengths] #kms
        
        # Separate Ly-alpha
        lya_ix = np.argmin(np.abs(self.emline_wavelengths - 1216.8))
        self.emline_sigma[lya_ix] = lya_sigma
        
        # Line EWs computed in `narrow_emission_lines`
        self.emline_eqw = [-1e10 for w in self.emline_wavelengths]
        
        # Emission line names
        waves, names = fsps_line_info()
        if np.allclose(self.emline_wavelengths, waves, 0.5):
            self.emline_names = names
        else:
            self.emline_names = ['?'] * len(self.emline_wavelengths)
            for w, n in zip(waves, names):
                dl = np.abs(self.emline_wavelengths - w)
                if dl.min() < 0.5:
                    self.emline_names[np.argmin(dl)] = n
        
        for l in self.emline_names:
            self.scale_lines[l] = 1.
            
        # Precomputed arrays for WG00 reddening defined between 0.1..3 um
        self.wg00lim = (self.wavelengths > 1000) & (self.wavelengths < 3.e4)
        self.wg00red = (self.wavelengths > 1000)*1.
        
        self.exec_params = None
        self.narrow = None
        
    def narrow_emission_lines(self, tage=0.1, emwave=DEFAULT_LINES, line_sigma=100, oversample=5, clip_sigma=10, verbose=False, get_eqw=True, scale_lyman_series=None, scale_lines={}, force_recompute=False, use_sigma_smooth=True, lorentz=False, **kwargs):
        """
        Replace broad FSPS lines with specified line widths
    
        tage : age in Gyr of FSPS model
        FSPS sigma: line width in A in FSPS models
        emwave : (approx) wavelength of line to replace
        line_sigma : line width in km/s of new line
        oversample : factor by which to sample the Gaussian profiles
        clip_sigma : sigmas from line center to use for the line
        scale_lyman_series : scaling to apply to Lyman-series emission lines
        scale_lines : scaling to apply to other emission lines, by name
        
        Returns: `dict` with keys
            wave_full, flux_full, line_full = wave and flux with fine lines
            wave, flux_line, flux_clean = original model + removed lines
            ymin, ymax = range of new line useful for plotting
        
        """
        if not hasattr(self, 'emline_dlam'):
            self._set_extend_attrs(line_sigma=line_sigma, **kwargs)
        
        self.params['add_neb_emission'] = True
        
        if scale_lyman_series is None:
            scale_lyman_series = self.scale_lyman_series
        else:
            self.scale_lyman_series = scale_lyman_series
        
        if scale_lines is None:
            scale_lines = self.scale_lines
        else:
            for k in scale_lines:
                if k in self.scale_lines:
                    self.scale_lines[k] = scale_lines[k]
                else:
                    print(f'Line "{k}" not found in `self.scale_lines`')
        
        # Avoid recomputing if all parameters are the same (i.e., change Av)
        call_params = np.hstack([self.param_floats(params=None), emwave, 
                                 list(self.scale_lines.values()), 
                        [tage, oversample, clip_sigma, scale_lyman_series]])
        try:
            is_close = np.allclose(call_params, self.exec_params)
        except:
            is_close = False
            
        if is_close & (not force_recompute):
             if verbose:
                 print('use stored')
             return self.narrow
        
        self.exec_params = call_params
        wave, line = self.get_spectrum(tage=tage, peraa=True)
        
        line_ix = [np.argmin(np.abs(self.emline_wavelengths - w)) 
                   for w in emwave]
                   
        line_lum = [self.emline_luminosity[i] for i in line_ix]
        line_wave = [self.emline_wavelengths[i] for i in line_ix]
        fsps_sigma = [np.sqrt((2*self.emline_dlam[i])**2 + 
             (self.params['sigma_smooth']/3.e5*self.emline_wavelengths[i])**2)
                              for i in line_ix]
        
        if line_sigma < 0:
            lines_sigma = [-line_sigma for ix in line_ix]
        elif (self.params['sigma_smooth'] > 0) & (use_sigma_smooth):
            lines_sigma = [self.params['sigma_smooth'] for ix in line_ix]
        else:
            lines_sigma = [self.emline_sigma[ix] for ix in line_ix]
            
        line_dlam = [sig/3.e5*lwave 
                     for sig, lwave in zip(lines_sigma, line_wave)]
    
        clean = line*1
        wlimits = [np.min(emwave), np.max(emwave)]
        wlimits = [2./3*wlimits[0], 4.3*wlimits[1]]
    
        wfine = utils.log_zgrid(wlimits, np.min(lines_sigma)/oversample/3.e5)
        qfine = wfine < 0
    
        if verbose:
            msg = 'Matched line: {0} [{1}], lum={2}'
            for i, ix in enumerate(line_ix):
                print(msg.format(line_wave[i], ix, line_lum[i]))
    
        ######### 
        # Remove lines from FSPS
        # line width seems to be 2*dlam at the line wavelength
        for i, ix in enumerate(line_ix):
            if self.params['nebemlineinspec']:
                gauss = 1/np.sqrt(2*np.pi*fsps_sigma[i]**2)
                gauss *= np.exp(-(wave - line_wave[i])**2/2/fsps_sigma[i]**2)
                clean -= gauss*line_lum[i]
            
            # indices of fine array where new lines defined
            qfine |= np.abs(wfine - line_wave[i]) < clip_sigma*line_dlam[i]
                
        # Linear interpolate cleaned spectrum on fine grid
        iclean = np.interp(wfine[qfine], wave, clean)
        
        # Append original and fine sampled arrays
        wfull = np.append(wave, wfine[qfine])
        cfull = np.append(clean, iclean)
        so = np.argsort(wfull)
        wfull, uniq = np.unique(wfull, return_index=True)
        cfull = cfull[uniq]
    
        gfull = cfull*0.
        for i in range(len(line_wave)):
                    
            if lorentz:
                # astropy.modeling.functional_models.Lorentz1D.html
                # gamma is FWHM/2., integral is gamma*pi
                gam = 2.35482*line_dlam[i]/2.
                gline = gam**2/(gam**2 + (wfull-line_wave[i])**2) 
                norm = line_lum[i]/(gam*np.pi)
            else:
                # Gaussian
                gline = np.exp(-(wfull - line_wave[i])**2/2/line_dlam[i]**2)
                norm = line_lum[i]/np.sqrt(2*np.pi*line_dlam[i]**2)
                
            if self.emline_names[line_ix[i]].startswith('Ly'):
                norm *= scale_lyman_series
            
            if self.emline_names[line_ix[i]] in self.scale_lines:
                norm *= self.scale_lines[self.emline_names[line_ix[i]]]
                    
            gfull += gline*norm
            
            if get_eqw:
                clip = np.abs(wfull - line_wave[i]) < clip_sigma*line_dlam[i]
                eqw = np.trapz(gline[clip]*norm/cfull[clip], wfull[clip])
                self.emline_eqw[line_ix[i]] = eqw
                
        cfull += gfull
    
        # For plot range
        ymin = iclean.min()
        line_peak = [1/np.sqrt(2*np.pi*dlam**2)*lum for dlam,lum in zip(line_dlam, line_lum)]
        ymax = iclean.max()+np.max(line_peak)
        
        data = OrderedDict()
        data['wave_full'] =  wfull
        data['flux_full'] =  cfull
        data['line_full'] =  gfull
        data['wave' ] =  wave
        data['flux_line' ] =  line
        data['flux_clean'] =  clean
        data['ymin' ] =  ymin
        data['ymax' ] =  ymax
        self.narrow = data
        
        return data


    def set_fir_template(self, arrays=None, file='templates/magdis/magdis_09.txt', verbose=True, unset=False, scale_pah3=0.5):
        """
        Set the far-IR template for reprocessed dust emission
        """
        if file.startswith('templates') & (not os.path.exists('templates')):
            file = os.path.join(utils.DATA_PATH, file)
        
        if unset:
            if verbose:
                print('Unset FIR template attributes')
            
            for attr in ['fir_template', 'fir_name', 'fir_arrays']:
                if hasattr(self, attr):
                    delattr(self, attr)
            
            return True
            
        if os.path.exists(file):
            if verbose:
                print(f'Set FIR dust template from {file}')
            _ = np.loadtxt(file, unpack=True)
            wave, flux = _[0], _[1]
            
            self.fir_name = file

        elif arrays is not None:
            if verbose:
                print(f'Set FIR dust template from input arrays')
            wave, flux = arrays
            self.fir_name = 'user-supplied'
        else:
            if verbose:
                print(f'Set FIR dust template from FSPS (DL07)')

            # Set with fsps
            self.params['dust1'] = 0
            self.params['dust2'] = 1.
            
            self.params['add_dust_emission'] = True
            wave, flux = self.get_spectrum(tage=1., peraa=True)
            self.params['add_dust_emission'] = False
            wave, flux_nodust = self.get_spectrum(tage=1., peraa=True)
            flux -= flux_nodust

            self.fir_name = 'fsps-dl07'
        
        if scale_pah3 is not None:
            if verbose:
                print(f'Scale 3.3um PAH: {scale_pah3:.2f}')
                
            ran = np.abs(wave-3.3e4) < 0.5e4
            line = np.abs(wave-3.3e4) < 0.3e4
            px = np.polyfit(wave[ran & ~line], flux[ran & ~line], 3)
            
            scaled_line = (flux[ran] - np.polyval(px, wave[ran]))*scale_pah3
            
            flux[ran] = np.polyval(px, wave[ran]) + scaled_line
                
        fir_flux = np.interp(self.wavelengths, wave, flux, left=0, right=0)
        self.fir_template = fir_flux/np.trapz(fir_flux, self.wavelengths)
        self.fir_arrays = arrays
        return True


    def set_dust(self, Av=0., dust_obj_type='WG00x', wg00_kwargs=WG00_DEFAULTS):
        """
        Set `dust_obj` attribute
        
        dust_obj_type: 
        
            'WG00'  = `dust_attenuation.radiative_transfer.WG00`
            'C00'   = `dust_attenuation.averages.C00`
            'WG00x' = `ParameterizedWG00`
            'KC13'  = Kriek & Conroy (2013) with dust_index parameter
            'R15'  = Reddy et al. (2015) with dust bump parameter
            
        ir_template: (wave, flux)
            Template to use for re-emitted IR light
            
        """
        from dust_attenuation import averages, radiative_transfer
        
        self.params['dust1'] = 0.
        self.params['dust2'] = 0.
                
        needs_init = False
        if not hasattr(self, 'dust_obj'):
            needs_init = True
        
        if hasattr(self, 'dust_obj_type'):
            if self.dust_obj_type != dust_obj_type:
                needs_init = True
        
        if needs_init:
            self.dust_obj_type = dust_obj_type
            if dust_obj_type == 'WG00':
                self.dust_obj = radiative_transfer.WG00(tau_V=1.0,
                                                        **WG00_DEFAULTS) 
            elif dust_obj_type == 'WG00x':
                self.dust_obj = ParameterizedWG00(Av=Av)             
            elif dust_obj_type == 'C00':
                self.dust_obj = averages.C00(Av=Av)
            elif dust_obj_type == 'R15':
                self.dust_obj = Reddy15(Av=Av, bump_ampl=2.)
            elif hasattr(dust_obj_type, 'extinguish'):
                self.dust_obj = dust_obj_type
            else:
                self.dust_obj = KC13(Av=Av)
                
            print('Init dust_obj: {0} {1}'.format(dust_obj_type, self.dust_obj.param_names))
            
        self.Av = Av
        
        if dust_obj_type == 'WG00':
            Avs = np.array([0.151, 0.298, 0.44 , 0.574, 0.825, 1.05 , 1.252, 1.428, 1.584, 1.726, 1.853, 1.961, 2.065, 2.154, 2.318, 2.454, 2.573, 2.686, 3.11 , 3.447, 3.758, 4.049, 4.317, 4.59 , 4.868, 5.148])
            taus = np.array([ 0.25,  0.5 ,  0.75,  1.  ,  1.5 ,  2.  ,  2.5 ,  3.  ,  3.5 , 4.  ,  4.5 ,  5.  ,  5.5 ,  6.  ,  7.  ,  8.  ,  9.  , 10.  , 15.  , 20.  , 25.  , 30.  , 35.  , 40.  , 45.  , 50.  ])
            tau_V = np.interp(Av, Avs, taus, left=0.25, right=50)
            self.dust_obj.tau_V = tau_V
            self.Av = self.dust_obj(5500*u.Angstrom)
        elif dust_obj_type == 'KC13':
            self.dust_obj.Av = Av
            self.dust_obj.delta = self.params['dust_index']
        else:
            self.dust_obj.Av = Av
    
    def redden(self, wave):
        if hasattr(self.dust_obj, 'extinguish'):
            return self.dust_obj.extinguish(wave, Av=self.Av)
        else:
            return 10**(-0.4*self.dust_obj(wave))
            
    def get_full_spectrum(self, tage=1.0, Av=0., get_template=True, set_all_templates=False, z=None, tie_metallicity=True, **kwargs):
        """
        Get full spectrum with reprocessed emission lines and dust emission
        
        dust_fraction: Fraction of the SED that sees the specified Av
        
        """
        
        # Set the dust model
        if Av is None:
            Av = self.Av
            
        if 'dust_obj_type' in kwargs:
            self.set_dust(Av=Av, dust_obj_type=kwargs['dust_obj_type'])
        elif hasattr(self, 'dust_obj'):
            self.set_dust(Av=Av, dust_obj_type=self.dust_obj_type)
        else:
            self.set_dust(Av=Av, dust_obj_type='WG00x')
        
        # Lognormal SFH?
        if ('lognorm_center' in kwargs) | ('lognorm_logwidth' in kwargs):
            self.set_lognormal_sfh(**kwargs)
        
        if 'lognorm_fburst' in kwargs:
            self.lognorm_fburst = kwargs['lognorm_fburst']
        
        # Burst fraction for lognormal SFH
        if self.is_lognorm_sfh:
            if not hasattr(self, '_lognorm_sfh'):
                self.set_lognormal_sfh()
        
            t1 = self.lognormal_integral(tage)
            dt = (tage-self._lognorm_sfh[0])
            t100 = (dt <= 0.1) & (dt >= 0)
            
            sfhy = self._lognorm_sfh[1]*1.
            sfhy += t1*10**self.lognorm_fburst/100e6*t100
            self.set_tabular_sfh(self._lognorm_sfh[0], sfhy) 
                             
        # Set FSPS parameters
        for k in kwargs:
            if k in self.params.all_params:
                self.params[k] = kwargs[k]
        
        if 'zmet' not in kwargs:
            self.params['zmet'] = self.izmet
        
        if ('gas_logz' not in kwargs) & tie_metallicity:
            self.params['gas_logz'] = self.params['logzsol']
            
        # Run the emission line function
        if tage is None:
            tage = self.params['tage']
            
        _ = self.narrow_emission_lines(tage=tage, **kwargs)

        wave = _['wave_full']
        flux = _['flux_full']
        lines = _['line_full']
        contin = flux - lines
        
        #self.sfr100 = self.sfr_avg(dt=np.minimum(tage, 0.1))
        
        # Apply dust
        if self.dust_obj_type == 'WG00':
            
            # To template
            red = (wave > 1000)*1.
            wlim = (wave > 1000) & (wave < 3.e4)
            red[wlim] = self.redden(wave[wlim]*u.Angstrom)
            
            # To lines
            red_lines = (self.emline_wavelengths > 1000)*1.
            wlim = (self.emline_wavelengths > 1000) 
            wlim &= (self.emline_wavelengths < 3.e4)
            line_wave = self.emline_wavelengths[wlim]*u.Angstrom
            red_lines[wlim] = self.redden(line_wave)
            
        else:
            red = self.redden(wave*u.Angstrom)
            
            if self.line_av_func is None:
                self.Av_line = self.Av*1.
                red_lines_full = red
                line_wave = self.emline_wavelengths*u.Angstrom
                red_lines = self.redden(line_wave)
            else:
                # Differential reddening towards nebular lines
                self.Av_line = self.line_av_func(Av)
                self.set_dust(Av=self.Av_line,
                              dust_obj_type=self.dust_obj_type)
                
                red_lines_full = self.redden(wave*u.Angstrom)
                line_wave = self.emline_wavelengths*u.Angstrom
                red_lines = self.redden(line_wave)
                
                # Reset for continuum
                self.set_dust(Av=Av, dust_obj_type=self.dust_obj_type)
                
        # Apply dust to line luminosities
        lred = [llum*lr for llum, lr in 
                        zip(self.emline_luminosity, red_lines)]
        self.emline_reddened = np.array(lred)
        
        # Total energy
        e0 = np.trapz(flux, wave)
        # Energy of reddened template
        
        reddened = contin*red+lines*red_lines_full
        e1 = np.trapz(reddened, wave)
        self.energy_absorbed = (e0 - e1)
                
        # Add dust emission
        if hasattr(self, 'fir_template') & self.params['add_dust_emission']:
            dust_em = np.interp(wave, self.wavelengths, self.fir_template)
            dust_em *= self.energy_absorbed
        else:
            dust_em = 0.
        
        meta0 = self.meta
        self.templ = self.as_template(wave, reddened+dust_em, meta=meta0)
        
        # Set template attributes
        if set_all_templates:
            
            # Original wavelength grid
            # owave = self.wavelengths
            # owave = owave[self.wg00lim]*u.Angstrom
            # self.wg00red[self.wg00lim] = self.redden(owave)
            # ofir = self.fir_template*self.energy_absorbed
            # fl_orig = _['flux_line']*self.wg00red + ofir
            # self.templ_orig = self.as_template(owave, fl_orig, meta=meta0)
            
            # No lines
            meta = meta0.copy()
            meta['add_neb_emission'] = False
            fl_cont = contin*red + dust_em
            #ocont = _['flux_clean']*self.wg00red + ofir
            self.templ_cont = self.as_template(wave, fl_cont, meta=meta)
            #self.templ_cont_orig = self.as_template(owave, ocont, meta=meta)
            
            # No dust
            meta = meta0.copy()
            meta['add_neb_emission'] = True
            meta['Av'] = 0
            self.templ_unred = self.as_template(wave, flux, meta=meta)
            #self.templ_unred_orig = self.as_template(owave, _['flux_clean'],
            #                                         meta=meta)
            
        if get_template:
            return self.templ
        else:
            return self.templ.wave, self.templ.flux
            
    def as_template(self, wave, flux, label=DEFAULT_LABEL, meta=None):
        """
        Return a `eazy.templates.Template` object with metadata
        """
        if meta is None:
            meta = self.meta
            
        templ = templates.Template(arrays=(wave, flux), meta=meta,
                                   name=label.format(**meta))
        return templ
    
    def lognorm_avg_sfr(self, tage=None, dt=0.1):
        """
        Analytic average SFR for lognorm SFH
        """
        if tage is None:
            tage = self.params['tage']
            
        t1 = self.lognormal_integral(tage)
        t0 = self.lognormal_integral(np.maximum(tage-dt, 0))
        sfr_avg = (t1*(1+10**self.lognorm_fburst)-t0)/(dt*1.e9)
        return sfr_avg
        
    @property 
    def sfr100(self):
        """
        SFR averaged over maximum(tage, 100 Myr) from `sfr_avg`
        """
        if self.params['sfh'] == 0:
            sfr_avg = 0.
        elif self.params['sfh'] == 3:
            # Try to integrate SFH arrays if attribute set
            if self.is_lognorm_sfh:
                sfr_avg = self.lognorm_avg_sfr(tage=None, dt=0.1)
                
            elif hasattr(self, '_sfh_tab'):
                try:
                    fwd = self.params['tabsfh_forward']
                except:
                    fwd = 1

                if fwd == 1:
                    age_lb = self.params['tage'] - self._sfh_tab[0]
                    step = -1
                else:
                    age_lb = self._sfh_tab[0]
                    step = 1
                    
                age100 = (age_lb <= 0.1) & (age_lb >= 0)
                if age100.sum() < 2:
                    sfr_avg = 0.
                else:
                    sfr_avg = np.trapz(self._sfh_tab[1][age100][::step],
                                       age_lb[age100][::step])/0.1
                
            else:
                sfr_avg = 0.
        else:
            sfr_avg = self.sfr_avg(dt=np.minimum(self.params['tage'], 0.1))
            
        return sfr_avg
    
    @property 
    def sfr10(self):
        """
        SFR averaged over last MAXIMUM(tage, 10 Myr) from `sfr_avg`
        """
        if self.params['sfh'] == 0:
            sfr_avg = 0.
        elif self.params['sfh'] == 3:
            # Try to integrate SFH arrays if attribute set
            if self.is_lognorm_sfh:
                sfr_avg = self.lognorm_avg_sfr(tage=None, dt=0.01)
                
            elif hasattr(self, '_sfh_tab'):
                try:
                    fwd = self.params['tabsfh_forward']
                except:
                    fwd = 1
                    
                if fwd == 1:
                    age_lb = self.params['tage'] - self._sfh_tab[0]
                    step = -1
                else:
                    age_lb = self._sfh_tab[0]
                    step = 1

                age10 = (age_lb < 0.01) & (age_lb >= 0)
                if age10.sum() < 2:
                    sfr_avg = 0.
                else:
                    sfr_avg = np.trapz(self._sfh_tab[1][age10][::step],
                                       age_lb[age10][::step])/0.1
                
            else:
                sfr_avg = 0.
        else:
            sfr_avg = self.sfr_avg(dt=np.minimum(self.params['tage'], 0.01))
            
        return sfr_avg
        
    @property 
    def meta(self):
        """
        Full metadata, including line properties
        """
        import fsps
        meta = self.param_dict
        
        if self._zcontinuous:
            meta['metallicity'] = 10**self.params['logzsol']*0.019
        else:
            meta['metallicity'] = self.zlegend[self.params['zmet']]
            
        for k in ['log_age','stellar_mass', 'formed_mass', 'log_lbol', 
                  'sfr', 'sfr100', 'dust_obj_type','Av','energy_absorbed', 
                  'fir_name', '_zcontinuous', 'scale_lyman_series',
                  'lognorm_center', 'lognorm_logwidth', 'is_lognorm_sfh', 
                  'lognorm_fburst']:
            if hasattr(self, k):
                meta[k] = self.__getattribute__(k)
        
        if hasattr(self, 'emline_names'):
            has_red = hasattr(self, 'emline_reddened')
            
            if self.emline_luminosity.ndim == 1:
                for i in range(len(self.emline_wavelengths)):
                    n = self.emline_names[i]
                    if n in self.scale_lines:
                        kscl = self.scale_lines[n]
                    else:
                        kscl = 1.0
                
                    meta[f'scale {n}'] = kscl
                    meta[f'line {n}'] = self.emline_luminosity[i]*kscl
                    if has_red:
                        meta[f'rline {n}'] = self.emline_reddened[i]*kscl
                    
                    meta[f'eqw {n}'] = self.emline_eqw[i]
                    meta[f'sigma {n}'] = self.emline_sigma[i]

        # Band information
        if hasattr(self, '_meta_bands'):
            light_ages = self.light_age_band(self._meta_bands, flat=False)
            band_flux = self.get_mags(tage=self.params['tage'], zmet=None,
                                      bands=self._meta_bands, units='flam')
            
            band_waves = [fsps.get_filter(b).lambda_eff*u.Angstrom 
                          for b in self._meta_bands]
            band_lum = [f*w for f, w in zip(band_flux, band_waves)]
            
            for i, b in enumerate(self._meta_bands):
                meta[f'lwage_{b}'] = light_ages[i]
                meta[f'lum_{b}'] = band_lum[i].value
        
        try:
            meta['libraries'] = ';'.join([s.decode() for s in self.libraries])  
        except:
            try:
                meta['libraries'] = ';'.join([s for s in self.libraries])  
            except:
                meta['libraries'] = '[error]'
                
        return meta
        
    @property 
    def param_dict(self):
        """
        `dict` version of `self.params`
        """
        d = OrderedDict()
        for p in self.params.all_params:
            d[p] = self.params[p]
        
        return d
    
    def light_age_band(self, bands=['v'], flat=True):
        """
        Get light-weighted age of current model
        """
        self.params['compute_light_ages'] = True
        band_ages = self.get_mags(tage=self.params['tage'], zmet=None, 
                                  bands=bands)
        self.params['compute_light_ages'] = False
        if flat & (band_ages.shape == (1,)):
            return band_ages[0]
        else:
            return band_ages
        
    def pset(self, params):
        """
        Return a subset dictionary of `self.meta`
        """
        d = OrderedDict()
        for p in params:
            if p in self.meta:
                d[p] = self.meta[p]
            else:
                d[p] = np.nan
                
        return d
        
    def param_floats(self, params=None):
        """
        Return a list of parameter values.  If `params` is None, then use 
        full list in `self.params.all_params`. 
        """
        
        if params is None:
            params = self.params.all_params
            
        d = []
        for p in params:
            d.append(self.params[p]*1)
        
        return np.array(d)
    
    def parameter_bounds(self, params, limit_age=False):
        """
        Parameter bounds for `scipy.optimize.least_squares`
        
        """
        blo = []
        bhi = []
        steps = []
        for p in params:
            if p in BOUNDS:
                blo.append(BOUNDS[p][0])
                bhi.append(BOUNDS[p][1])
                steps.append(BOUNDS[p][2])
            else:
                blo.append(-np.inf)
                bhi.append(np.inf)
                steps.append(0.05)
                
        return (blo, bhi), steps
            
    def fit_spec(self, wave_obs, flux_obs, err_obs, mask=None, plist=['tage', 'Av', 'gas_logu', 'sigma_smooth'], func_kwargs={'lorentz':False}, verbose=True, bspl_kwargs=None, cheb_kwargs=None, lsq_kwargs={'method':'trf', 'max_nfev':200, 'loss':'huber', 'x_scale':1.0, 'verbose':True}, show=False):
        """
        Fit models to observed spectrum
        """
        from scipy.optimize import least_squares
        import matplotlib.pyplot as plt
        
        import grizli.utils
        
        sys_err = 0.015
        
        if wave_obs is None:
            # mpdaf muse spectrum
            _ = None # muse spectrum object
            spec = _[0].spectra['MUSE_TOT_SKYSUB']
            wave_obs = spec.wave.coord()
            flux_obs = spec.data.filled(fill_value=np.nan)
            err_obs = np.sqrt(spec.var.filled(fill_value=np.nan))
            err_obs = np.sqrt(err_obs**2+(sys_err*flux_obs)**2) 
            
            mask = np.isfinite(flux_obs+err_obs) & (err_obs > 0) 
            omask = mask
            
            #mask = omask & (wave_obs/(1+0.0342) > 6520) & (wave_obs/(1+0.0342) < 6780)
            mask = omask & (wave_obs/(1+0.0342) > 4800) & (wave_obs/(1+0.0342) < 5050)
            
        theta0 = np.array([self.meta[p] for p in plist])
        
        if bspl_kwargs is not None:
            bspl = grizli.utils.bspline_templates(wave_obs, get_matrix=True, 
                                                  **bspl_kwargs)
        elif cheb_kwargs is not None:
            bspl = grizli.utils.cheb_templates(wave_obs, get_matrix=True, 
                                                  **cheb_kwargs)
        else:
            bspl = None
            
        kwargs = func_kwargs.copy()
        for i, p in enumerate(plist):
            kwargs[p] = theta0[i]
        
        # Model test
        margs = (self, plist, wave_obs, flux_obs, err_obs,
                 mask, bspl, kwargs, 'model')
        
        flux_model, Anorm, chi2_init = self.objfun_fitspec(theta0, *margs)
        
        if show:                
            mask &= np.isfinite(flux_model+flux_obs+err_obs) & (err_obs > 0) 

            plt.close('all')
            
            fig = plt.figure(figsize=(12, 6))
            plt.errorbar(wave_obs[mask],
                         flux_obs[mask],
                         err_obs[mask],
                         color='k',
                         alpha=0.5,
                         linestyle='None',
                         marker='.')
                         
            plt.plot(wave_obs, flux_model, color='pink', linewidth=2, alpha=0.8)
        else:
            fig = None
            
        bounds, steps = self.parameter_bounds(plist)
        #lsq_kwargs['diff_step'] = np.array(steps)/2.
        #lsq_kwargs['diff_step'] = 0.05
        lsq_kwargs['diff_step'] = steps
        lmargs = (self, plist, wave_obs, flux_obs, err_obs,
                  mask, bspl, kwargs, 'least_squares verbose')
        
        _res = least_squares(self.objfun_fitspec,
                             theta0,bounds=bounds,
                             args=lmargs,
                             **lsq_kwargs,
                             )
        
        fit_model, Anorm, chi2_fit = self.objfun_fitspec(_res.x, *margs)
        
        result = {'fit_model':fit_model, 'Anorm':Anorm, 'chi2_fit':chi2_fit, 
                  'least_squares':_res, 'bounds':bounds, 'plist':plist, 
                  'lsq_kwargs':lsq_kwargs, 'bspl':bspl}
        
        return result


    @staticmethod
    def objfun_fitspec(theta, self, plist, wave_obs, flux_obs, err_obs, mask, bspl, kwargs, ret_type):
        """
        Objective function for fitting spectra
        """
        import scipy.stats
        
        try:
            from grizli.utils_c.interp import interp_conserve_c as interp_func
        except:
            interp_func = utils.interp_conserve
        
        err_scale = 1.
            
        for i, p in enumerate(plist):
            if p == 'err_scale':
                err_scale = theta[i]
                continue
                
            kwargs[p] = theta[i]
                    
        templ = self.get_full_spectrum(**kwargs)
                
        flux_model = templ.to_observed_frame(z=self.params['zred'],
                                             lsf_func=self.lsf_func,
                                             clip_wavelengths=None,
                                             wavelengths=wave_obs,
                                             smoothspec_kwargs={'fftsmooth':self.FFT_SMOOTH},
                                             )
        # flux_model = templ.resample(wave_obs, z=self.params['zred'],
        #                            in_place=False,
        #                            return_array=True, interp_func=interp_func)
        
        flux_model = np.squeeze(flux_model.flux_flam())
        
        if mask is None:
            mask = np.isfinite(flux_model+flux_obs+err_obs) & (err_obs > 0)
        
        if bspl is not None:
            _A = (bspl.T*flux_model)
            _yx = (flux_obs / err_obs)[mask]
            _c = np.linalg.lstsq((_A/err_obs).T[mask,:], _yx, rcond=-1)
            Anorm = np.mean(bspl.dot(_c[0]))
            flux_model = _A.T.dot(_c[0])
            
        else:
            lsq_num = (flux_obs*flux_model/err_obs**2)[mask].sum()
            lsq_den = (flux_model**2/err_obs**2)[mask].sum()
            Anorm = lsq_num/lsq_den
            flux_model *= Anorm
        
        chi = ((flux_model - flux_obs)/err_obs)[mask]
        chi2 = (chi**2).sum()
                                       
        if 'verbose' in ret_type:
            print('{0} {1:.4f}'.format(theta, chi2))
            
        if 'model' in ret_type:
            return flux_model, Anorm, chi2
        elif 'least_squares' in ret_type:
            return chi
        elif 'logpdf' in ret_type:
            #return -chi2/2.
            lnp = scipy.stats.norm.logpdf((flux_model-flux_obs)[mask], 
                                           loc=0, 
                                           scale=(err_obs*flux_model)[mask])
            
            return lnp
        else:
            return chi2
    
    def line_to_obsframe(self, zred=None, cosmology=None, verbose=False, unit=LINE_CGS, target_stellar_mass=None, target_sfr=None, target_lir=None):
        """
        Scale factor to convert internal line luminosities (L_sun) to observed frame
        
        If ``target_stellar_mass``, ``target_sfr``, or ``target_lir`` 
        specified, then scale the output to the desired value using the 
        intrinsic properties.  Units are linear ``Msun``, ``Msun/yr``, 
        and ``Lsun``, respectively.
        
        """
        from astropy.constants import L_sun
        
        if zred == None:
            zred = self.params['zred']
            if verbose:
                msg = 'continuum_to_obsframe: Use params[zred] = {0:.3f}'
                print(msg.format(zred))
                    
        if cosmology is None:
            cosmology = self.cosmology
        else:
            self.cosmology = cosmology
        
        if zred <= 0:
            dL = 1*u.cm
        else:
            dL = cosmology.luminosity_distance(zred).to(u.cm)

        to_cgs = (1*L_sun/(4*np.pi*dL**2)).to(unit)
        
        if target_stellar_mass is not None:
            to_cgs *= target_stellar_mass / self.stellar_mass
        elif target_sfr is not None:
            to_cgs *= target_sfr / self.sfr100
        elif target_lir is not None:
            to_cgs *= target_lir / self.energy_absorbed
        
        return to_cgs.value
        
    def continuum_to_obsframe(self, zred=None, cosmology=None, unit=u.microJansky, verbose=False, target_stellar_mass=None, target_sfr=None, target_lir=None):
        """
        Compute a normalization factor to scale input FSPS model flux density 
        units of (L_sun / Hz) or (L_sun / \AA) to observed-frame `unit`.
        
        If ``target_stellar_mass``, ``target_sfr``, or ``target_lir`` 
        specified, then scale the output to the desired value using the 
        intrinsic properties.  Units are linear ``Msun``, ``Msun/yr``, 
        and ``Lsun``, respectively.
        """
        from astropy.constants import L_sun

        if zred == None:
            zred = self.params['zred']
            if verbose:
                msg = 'continuum_to_obsframe: Use params[zred] = {0:.3f}'
                print(msg.format(zred))
                    
        if cosmology is None:
            cosmology = self.cosmology
        else:
            self.cosmology = cosmology
        
        if zred <= 0:
            dL = 1*u.cm
        else:
            dL = cosmology.luminosity_distance(zred).to(u.cm)
        
        # FSPS L_sun / Hz to observed-frame
        try:
            # Unit is like f-lambda
            _x = (1*unit).to(u.erg/u.second/u.cm**2/u.Angstrom) 
            is_flam = True
            obs_unit = (1*L_sun/u.Angstrom/(4*np.pi*dL**2)).to(unit)/(1+zred)
        except:
            # Unit is like f-nu
            is_flam = False
            obs_unit = (1*L_sun/u.Hz/(4*np.pi*dL**2)).to(unit)*(1+zred)
        
        if target_stellar_mass is not None:
            obs_unit *= target_stellar_mass / self.stellar_mass
        elif target_sfr is not None:
            obs_unit *= target_sfr / self.sfr100
        elif target_lir is not None:
            obs_unit *= target_lir / self.energy_absorbed
            
        return obs_unit.value
        
    def fit_phot(self, phot_dict, filters=None, flux_unit=u.microJansky, plist=['tage', 'Av', 'gas_logu', 'sigma_smooth'], func_kwargs={'lorentz':False}, verbose=True, lsq_kwargs={'method':'trf', 'max_nfev':200, 'loss':'huber', 'x_scale':1.0, 'verbose':True}, show=False, TEF=None, photoz_obj=None):
        """
        Fit models to observed spectrum
        """
        from scipy.optimize import least_squares
        import matplotlib.pyplot as plt
        import grizli.utils
        
        sys_err = 0.02
        
        flux = phot_dict['fobs']
        err = phot_dict['efobs']
        if 'flux_unit' in phot_dict:
            flux_unit = phot_dict['flux_unit']
            
        x0 = np.array([self.meta[p] for p in plist])
        
        # Are input fluxes f-lambda or f-nu?
        try:
            _x = (1*flux_unit).to(u.erg/u.second/u.cm**2/u.Angstrom) 
            is_flam = True
        except:
            is_flam = False
         
        # Initialize keywords   
        kwargs = func_kwargs.copy()
        for i_p, p in enumerate(plist):
            kwargs[p] = x0[i_p]
        
        # Initial model
        margs = (self, plist, flux, err, is_flam, filters, TEF, kwargs, 'model')        
        flux_model, Anorm, chi2_init, templ = self.objfun_fitphot(x0, *margs)
        
        if show:                
            
            if hasattr(show, 'plot'):
                ax = show
            else:
                plt.close('all')
                fig = plt.figure(figsize=(12, 6))
                ax = fig.add_subplot(111)
                
            mask = err > 0
            pivot = np.array([f.pivot for f in filters])
            so = np.argsort(pivot)
            
            ax.errorbar(pivot[mask]/1.e4, flux[mask], err[mask],
                        color='k', alpha=0.5, linestyle='None', marker='.')
            ax.scatter(pivot[so]/1.e4, flux_model[so], color='pink', 
                       alpha=0.8, zorder=100)
        
        # Parameter bounds    
        bounds, steps = self.parameter_bounds(plist)
        lsq_kwargs['diff_step'] = steps
        
        # Run the optimization
        lmargs = (self, plist, flux, err, is_flam, filters, TEF, kwargs, 'least_squares verbose')        
        _res = least_squares(self.objfun_fitphot, x0, bounds=bounds,
                             args=lmargs, **lsq_kwargs)
        
        _out = self.objfun_fitphot(_res.x, *margs)
        
        xtempl = _out[3]
        xscale = _out[1]
        
        _fit = {}
        _fit['fmodel'] = _out[0]
        _fit['scale'] = xscale
        _fit['chi2'] = _out[2]
        _fit['templ'] = xtempl
        _fit['plist'] = plist
        _fit['theta'] = _res.x
        _fit['res'] = _res
        
        # Stellar mass
        #fit_model, Anorm, chi2_fit, templ = _phot
        
        # Parameter scaling to observed frame.  
        # e.g., stellar mass = self.stellar_mass * scale / to_obsframe
        z = self.params['zred']
        _obsfr = self.continuum_to_obsframe(zred=z, unit=flux_unit)
        _fit['to_obsframe'] = _obsfr
        
        scl = _fit['scale']/_fit['to_obsframe']
        _fit['log_mass'] = np.log10(self.stellar_mass*scl)
        _fit['sfr'] = self.sfr*scl
        _fit['sfr10'] = self.sfr10*scl
        _fit['sfr100'] = self.sfr100*scl
        age_bands = ['i1500','v']
        ages = self.light_age_band(bands=age_bands)
        for i, b in enumerate(age_bands):
            _fit['age_'+b] = ages[i]
            
        if show:
            ax.scatter(pivot[so]/1.e4, _fit['fmodel'][so], color='r', 
                      alpha=0.8, zorder=101)
            
            iz = templ.zindex(z)
            igm = templ.igm_absorption(z, scale_tau=1.4)
            
            if is_flam:
                ax.plot(xtempl.wave*(1+z)/1.e4, xtempl.flux[iz,:]*xscale*igm, 
                    color='r', alpha=0.3, zorder=10000)
            else:
                ax.plot(xtempl.wave*(1+z)/1.e4, xtempl.flux_fnu(iz)*xscale*igm, 
                    color='r', alpha=0.3, zorder=10000)
        
        return _fit
        
    @staticmethod
    def objfun_fitphot(theta, self, plist, flux_fnu, err_fnu, is_flam, filters, TEF, kwargs, ret_type):
        """
        Objective function for fitting spectra
        """
        try:
            from grizli.utils_c.interp import interp_conserve_c as interp_func
        except:
            interp_func = utils.interp_conserve
            
        for i, p in enumerate(plist):
            kwargs[p] = theta[i]
                    
        templ = self.get_full_spectrum(**kwargs)
        model_fnu = templ.integrate_filter_list(filters,
                                     z=self.params['zred'], flam=is_flam, 
                                     include_igm=True)
        
        mask = (err_fnu > 0)
        full_var = err_fnu**2
        
        if TEF is not None:
            tefz = TEF(self.params['zred'])
            full_var += (flux_fnu*tefz)**2
            
        lsq_num = (flux_fnu*model_fnu/full_var)[mask].sum()
        lsq_den = (model_fnu**2/full_var)[mask].sum()
        Anorm = lsq_num/lsq_den
        model_fnu *= Anorm
        
        chi = ((model_fnu - flux_fnu)/np.sqrt(full_var))[mask]
        chi2 = (chi**2).sum()
        
        if 'verbose' in ret_type:
            print('{0} {1:.4f}'.format(theta, (chi**2).sum()))
            
        if 'model' in ret_type:
            return model_fnu, Anorm, chi2, templ
        elif 'least_squares' in ret_type:
            return chi
        elif 'logpdf' in ret_type:
            return -chi2/2
        else:
            return chi2


    def fit_grism(self, mb, plist=['tage', 'zred'], func_kwargs={'lorentz':False}, verbose=True, lsq_kwargs={'method':'trf', 'max_nfev':200, 'loss':'huber', 'x_scale':1.0, 'verbose':True}, show=False, TEF=None, photoz_obj=None, flux_unit=u.erg/u.second/u.cm**2/u.Angstrom):
        """
        Fit models to grism spectrum
        """
        from scipy.optimize import least_squares
        import matplotlib.pyplot as plt
        import grizli.utils

        x0 = np.array([self.meta[p] for p in plist])

        # Initialize keywords   
        kwargs = func_kwargs.copy()
        for i_p, p in enumerate(plist):
            kwargs[p] = x0[i_p]

        # Initial model
        margs = (self, plist, mb, TEF, kwargs, 'model')        
        tfit, Anorm, chi2_init, templ = self.objfun_fitgrism(x0, *margs)

        if show:                
            _init_fig = mb.oned_figure(tfit=tfit, show_individual_templates=True)

        # Parameter bounds    
        bounds, steps = self.parameter_bounds(plist)
        lsq_kwargs['diff_step'] = steps

        # Run the optimization
        lmargs = (self, plist, mb, TEF, kwargs, 'least_squares verbose')        
        _res = least_squares(self.objfun_fitgrism, x0, bounds=bounds,
                             args=lmargs, **lsq_kwargs)

        _out = self.objfun_fitgrism(_res.x, *margs)

        xtempl = _out[3]
        xscale = _out[1]

        _fit = {}
        _fit['fmodel'] = _out[0]
        _fit['scale'] = xscale
        _fit['chi2'] = _out[2]
        _fit['templ'] = xtempl
        _fit['plist'] = plist
        _fit['theta'] = _res.x
        _fit['res'] = _res

        # Stellar mass
        #fit_model, Anorm, chi2_fit, templ = _phot

        # Parameter scaling to observed frame.  
        # e.g., stellar mass = self.stellar_mass * scale / to_obsframe
        z = self.params['zred']
        _obsfr = self.continuum_to_obsframe(zred=z, unit=flux_unit)
        _fit['to_obsframe'] = _obsfr

        scl = _fit['scale']/_fit['to_obsframe']
        _fit['log_mass'] = np.log10(self.stellar_mass*scl)
        _fit['sfr'] = self.sfr*scl
        _fit['sfr10'] = self.sfr10*scl
        _fit['sfr100'] = self.sfr100*scl
        age_bands = ['i1500','v']
        ages = self.light_age_band(bands=age_bands)
        for i, b in enumerate(age_bands):
            _fit['age_'+b] = ages[i]

        _tfit = _out[0]

        if show:
            _fit_fig = mb.oned_figure(tfit=_tfit, show_individual_templates=True)
            _figs = (_init_fig, _fit_fig)
        else:
            _figs = None

        return _fit, _tfit, _figs

    @staticmethod
    def objfun_fitgrism(theta, self, plist, mb, TEF, kwargs, ret_type):
        """
        Objective function for fitting spectra
        """
        try:
            from grizli.utils_c.interp import interp_conserve_c as interp_func
        except:
            interp_func = utils.interp_conserve

        from grizli import utils

        for i, p in enumerate(plist):
            kwargs[p] = theta[i]

        templ = self.get_full_spectrum(**kwargs)
        tsp = utils.SpectrumTemplate(wave=templ.wave, flux=templ.flux_flam())
        tx = {'fsps':tsp}

        if 'least_squares' in ret_type:
            out = mb.xfit_at_z(z=self.params['zred'],
                               templates=tx, fitter='lstsq',
                               fit_background=True, get_uncertainties=False,
                               include_photometry=False, get_residuals=True,
                               use_cached_templates=False)

            chi, _coeffs, _, _ = out
            chi2 = (chi**2).sum()
            Anorm = _coeffs[-1]
        else:
            tfit = mb.template_at_z(z=self.params['zred'], templates=tx)
            chi2 = tfit['chi2']
            Anorm = tfit['cfit']['fsps'][0]

        if 'verbose' in ret_type:
            print('{0} {1:.4f}'.format(theta, chi2))

        if 'model' in ret_type:
            return tfit, Anorm, chi2, templ
        elif 'least_squares' in ret_type:
            return chi
        elif 'logpdf' in ret_type:
            return -chi2/2
        else:
            return chi2



import os
import collections
import numpy as np

from . import utils

__all__ = ["EazyParam", "TranslateFile", "read_param_file"]

def read_param_file(param_file=None, verbose=True):
    """
    Load a param file and add default parameters if any missing
    """
    param = EazyParam(param_file, verbose=True)
    if param_file is not None:
        # Read defaults
        defaults = EazyParam(None, verbose=False)
        for k in defaults.param_names:
            if k not in param.param_names:
                param[k] = defaults[k]
                if verbose:
                    print(f'Parameter default: {k} = {defaults[k]}')

    return param


class EazyParam(object):
    def __init__(self, PARAM_FILE=None, verbose=True):
        """
        Read an Eazy zphot.param file.

        Example: 

            >>> if os.path.exists('zphot.param'):
            ...     params = EazyParam(PARAM_FILE='zphot.param')
            ...     print(params['Z_STEP'])

        Defaults are in `eazy/data/zphot.param.default <https://github.com/gbrammer/eazy-py/blob/master/eazy/data/zphot.param.default>`_
        
        Parameters
        ----------
        param_file : str
            Name of parameter file.  If None, then will get 
            `data/zphot.param.default` from within the module.
        
        Attributes
        ----------
        params : `collections.OrderedDict`
            Parameter dictionary.  Don't modify this directly but rather use
            `__getitem__` and `__setitem__` methods.
        
        param_names
        
        formats : list
            List indicating if parameters are interpreted as string ('s') or 
            scalar ('f') values.  
            
        """
        if PARAM_FILE is None:
            PARAM_FILE = os.path.join(os.path.dirname(__file__), 
                                      'data/zphot.param.default')
        
        if verbose:
            print('Read default param file: '+PARAM_FILE)
            
        self.filename = PARAM_FILE
        self.param_path = os.path.dirname(PARAM_FILE)
           
        f = open(PARAM_FILE,'r')
        self.lines = f.readlines()
        f.close()

        self.params = collections.OrderedDict()
        self.formats = collections.OrderedDict()
        
        self._process_params()


    @property 
    def param_names(self):
        """
        Keywords of the `params` dictionary
        """
        return list(self.params.keys())


    def _process_params(self):
        """
        Process parameter dictionary
        """
        params = collections.OrderedDict()
        formats = collections.OrderedDict()
        
        #self.param_names = []
        for line in self.lines:
            if not line.strip().startswith('#'):
                lsplit = line.split()
                if lsplit.__len__() >= 2:
                    params[lsplit[0]] = lsplit[1]
                    #self.param_names.append(lsplit[0])
                    try:
                        flt = float(lsplit[1])
                        formats[lsplit[0]] = 'f'
                        params[lsplit[0]] = flt
                    except:
                        formats[lsplit[0]] = 's'
                    
        self.params = params
        self.formats = formats


    @property
    def to_mJy(self):
        """
        Return catalog conversion factor to mJy based on ``PRIOR_ABZP``.
        """
        return 10**(-0.4*(self.params['PRIOR_ABZP']-23.9))/1000.


    def write(self, file=None):
        """
        Write to an ascii file
        """
        if file == None:
            print('No output file specified...')
        else:
            fp = open(file,'w')
            for param in self.param_names:
                fp.write('{0:25s} {1}\n'.format(param, self.params[param]))

            fp.close()


    def __getitem__(self, param_name):
        """
        Get item from ``params`` dict and return None if parameter not found.
        """
        if param_name.upper() not in self.param_names:
            print(f'Parameter {param_name} not found.  Check `param_names`'              
                    ' attribute.')
            return None
        else:
            return self.params[param_name.upper()]


    def __setitem__(self, param_name, value):
        """
        Set item in ``params`` dict.
        """
        self.params[param_name.upper()] = value


    def verify_params(self):
        """
        Some checks on the parameters
        """
        
        assert(self['Z_MAX'] > self['Z_MIN'])
        
        test_paths = ['./', utils.DATA_PATH, os.path.join(utils.DATA_PATH, 'filters')]
        
        for k in ['TEMPLATES_FILE', 'TEMP_ERR_FILE', 'CATALOG_FILE', 
                  'FILTERS_RES']:
            if isinstance(self[k], str):
                filename = self[k]
                file_found = False
                for path in test_paths:
                    if os.path.exists(os.path.join(path, filename)):
                        file_found = True
                        break
                    
                if not file_found:
                    raise FileNotFoundError(
                        f'{k} ({self[k]}) not found in {test_paths}'
                    )

        assert(int(self['ARRAY_NBITS']) in [32,64])
        
        # Positive
        for k in ['TEMP_ERR_A2', 'SYS_ERR', 'IGM_SCALE_TAU', 'MW_EBV', 
                  'OMEGA_M', 'OMEGA_L']:
            if self[k] < 0:
                raise ValueError(f'{k} ({self[k]}) must be >= 0')
        
        # Positive nonzero
        for k in ['Z_STEP','H0', 'RF_PADDING']:
            if self[k] < 0:
                raise ValueError(f'{k} ({self[k]}) must be > 0')


    @property 
    def kwargs(self):
        """
        Dictionary with lower-case parameter names for passing as ``**kwargs``
        """
        kws = collections.OrderedDict()
        for k in self.param_names:
            kws[k.lower()] = self.params[k]
        
        return kws
        
        
class TranslateFile():
    def __init__(self, file='zphot.translate'):
        """
        File for translating catalog columns to associate bandbasses to them
        
        The `file` has format
                
        .. code-block::

            flux_irac_ch1  F18
            err_irac_ch1   E18
            ...
            
        or a CSV table with format
        
        .. code-block::
        
            column, trans, error
            flux_irac_ch1, F18
            err_irac_ch1,  E18, 1.0
            ...
        
        where `flux_irac_ch1` is a column in the catalog table corresponding 
        to the IRAC 3.6 µm flux density. ``F18`` indicates that this is a 
        *flux density* column and is associated with filter number 18 in the 
        `~eazy.params.filters.FilterFile`.
        
        ``E18`` indicates an uncertainty column, and filters must have both 
        flux density and uncertainty columns to be considered.
        
        The original catalog could have had column names ``F18`` and ``E18``
        and not needed a translate file but it is generally preferable to have
        more descriptive column names that aren't necessarily tied to a
        particular `eazy` filter file.
        
        Note, similarly, that columns like `F{N}` and `E{N}` are treated as 
        these types of flux and uncertainty columns.  If they correspond to 
        something else, they should be "translated" to avoid confusion
        
        """
        from astropy.table import Table
        
        self.file = file
        self.trans = collections.OrderedDict()
        self.error = collections.OrderedDict()

        if hasattr(file, 'colnames'):
            tr = file
            self.file = 'input_table.translate'
        
            if 'error' not in tr.colnames:
                tr['error'] = 1.0
        
            if tr.colnames != ['column', 'trans', 'error']:
                msg = f"table translate_file file must have columns"
                msg += f" 'column', 'trans' [, 'error'].  The file {file}"
                msg += f' has columns {tr.colnames}.'
                raise ValueError(msg)
        
            for i, k in enumerate(tr['column']):
                self.trans[k] = tr['trans'][i]
                self.error[k] = tr['error'][i]

        elif file.endswith('csv'):
            tr = Table.read(file)
                
            if 'error' not in tr.colnames:
                tr['error'] = 1.0
            
            if tr.colnames != ['column', 'trans', 'error']:
                msg = f"csv translate_file file must have columns"
                msg += f" 'column', 'trans' [, 'error'].  The file {file}"
                msg += f' has columns {tr.colnames}.'
                raise ValueError(msg)
            
            for i, k in enumerate(tr['column']):
                self.trans[k] = tr['trans'][i]
                self.error[k] = tr['error'][i]
                            
        else:
            lines = open(file).readlines()

            for line in lines:
                spl = line.split()
                if (line.strip() == '') | (len(spl) < 2):
                    continue
            
                key = spl[0]
                self.trans[key] = spl[1]
                if len(spl) == 3:
                    self.error[key] = float(spl[2])
                else:
                    self.error[key] = 1.


    def change_error(self, filter=88, value=1.e8):
        """
        Modify uncertainties based on error scaling factors in translate file
        """
        if isinstance(filter, str):
            if 'f_' in filter:
                err_filt = filter.replace('f_','e_')
            else:
                err_filt = 'e'+filter

            if err_filt in self.error:
                self.error[err_filt] = value
                return True
        
        if isinstance(filter, int):
            for key in self.trans.keys():
                if self.trans[key] == 'E{0:0d}'.format(filter):
                    self.error[key] = value
                    return True
        
        print('Filter {0} not found in list.'.format(str(filter)))


    def write(self, file=None, show_ones=False):
        """
        Write to an ascii file
        """
        lines = []
        for key in self.error:
            line = '{0}  {1}'.format(key, self.trans[key])
            if self.trans[key].startswith('E') & ((self.error[key] != 1.0) | show_ones):
                line += '  {0:.1f}'.format(self.error[key])

            lines.append(line+'\n')

        if file is None:
            file = self.file
        
        if file:
            fp = open(file,'w')
            fp.writelines(lines)
            fp.close()
        else:
            for line in lines:
                print(line[:-1])


    def to_csv(self):
        """
        Generate CSV string
        """
        rows = 'column,trans,error\n'
        for k in self.error:
            rows += f'{k},{self.trans[k]},{self.error[k]}\n'
        return rows
        


"""
Scripts for more interactive visualization of SEDs, etc.
"""
import numpy as np
import astropy.io.fits as pyfits
import astropy.wcs as pywcs

from . import utils

__all__ = ['EazyExplorer']

CUTOUT_URL = "https://grizli-cutout.herokuapp.com/thumb?all_filters=True&size=4&scl=1.0&asinh=True&filters=f115w-clear,f277w-clear,f444w-clear&rgb_scl=1.5,0.74,1.3&pl=2&ra={ra}&dec={dec}"

EXTRA_SLIDER_KWS = {'column':'id', 'label':'ID',
                    'min':0, 'max':10, 'step':1,
                    'value':[1,5], 'checked':[],
                    }

class EazyExplorer(object):
    def __init__(self, photoz, zout, extra_zout_columns=[], selection=None, 
                 extra_plots={}, extra_slider_kws=EXTRA_SLIDER_KWS):
        """
        Generating a tool for interactive visualization of `eazy` outputs with 
        the `dash` + `plotly` libraries.
        
        Parameters
        ----------
        photoz : `~eazy.photoz.PhotoZ`
            The main ``PhotoZ`` object.
        
        zout : `astropy.table.Table`
            The `zout` output table with galaxy parameters.  At
            a minimum, it must have columns ``id``, ``z_phot``, ``z_spec``, 
            ``z_phot_chi2``, ``nusefilt``, ``restU``, ``restV``, ``restJ``, 
            ``sfr``, ``mass``, ``ra``, ``dec``.
        
        extra_zout_columns : list
            Additional columns from ``zout`` to copy to the app 
            ``df`` `pandas.DataFrame` object.
        
        selection : array-like
            Selection array on `zout` for catalog subset, can be integer 
            indices, or a boolean array with the same length as `zout`
        
        extra_plots : dict
            Extra scatter plot definitions following
            
            >>> extra_plots = {'PlotName': (xcol, ycol, xlabel, ylabel, xr, yr)}
            
            where ``xcol`` and ``ycol`` are `str` column names in `zout`,
            ``xlabel`` and ``ylabel`` are `str` used for the plot labels and
            ``xr`` and ``yr`` are the default plot range tuples. Only
            ``xcol`` and ``ycol`` are required, and the others are computed if
            not provided.  Note that ``xcol`` and ``ycol`` are added 
            automatically from `zout` if necessary and don't need to be 
            specified in `extra_zout_columns`.
            
        """
        import pandas as pd
        from astropy.table import Table
        
        try:
            import dash
            from dash import  dcc, html
            import plotly.express as px
            
        except ImportError:
            print('Failed to import dash & plotly, so the interactive tool'
                  'won\'t work.\n'
                  'Install with `pip install "dash>=2.5.1"` and also '
                  '`pip install jupyter_dash` for running a server '
                  'through jupyter')
                  
        uv = -2.5*np.log10(zout['restU']/zout['restV'])
        vj = -2.5*np.log10(zout['restV']/zout['restJ'])
        ssfr = zout['sfr']/zout['mass']
        
        for c in ['z_phot', 'z_spec']:
            if hasattr(zout[c], 'mask'):
                zout[c].fill_value = -0.1
                
        def fill_masked(_data, fill_value=None):
            if hasattr(_data, 'mask'):
                if fill_value is None:
                    return _data.filled()
                else:
                    return _data.filled(fill_value)
            else:
                return _data
        
        self.extra_slider_kws = extra_slider_kws
        if 'flux_radius' in zout.colnames:
            self.extra_slider_kws = {'column':'flux_radius', 'label':'R50',
                                     'min':0, 'max':20, 'step':0.2,
                                     'value':[1,10], 'checked':[],
                                     }
                                
        #df = pd.DataFrame()
        df = Table()
        df['id'] = fill_masked(zout['id'])
        df['nusefilt'] = fill_masked(zout['nusefilt'])
        df['uv'] = fill_masked(uv)
        df['vj'] = fill_masked(vj)
        df['ssfr'] = fill_masked(np.log10(ssfr))
        df['mass'] = fill_masked(np.log10(zout['mass']))
        df['z_phot'] = fill_masked(zout['z_phot'])
        df['z_spec'] = np.clip(fill_masked(zout['z_spec']), -0.1, 12)
        df['ra'] = fill_masked(photoz.RA)
        df['dec'] = fill_masked(photoz.DEC)
        df['chi2'] = fill_masked(zout['z_phot_chi2']/zout['nusefilt'])
        
        self.zp = photoz.zp*1
        
        for c in extra_zout_columns:
            if c in zout.colnames:
                df[c] = fill_masked(zout[c])
        
        col = self.extra_slider_kws['column']
        if col in zout.colnames:
            if col not in df.columns:
                df[col] = fill_masked(zout[col])
        else:
            if col not in df.columns:
                df[col] = df['id']
        
        _red_ix = np.argmax(photoz.pivot*(photoz.pivot < 3.e4))
        self.DEFAULT_FILTER = photoz.flux_columns[_red_ix]
                
        ZP = photoz.param['PRIOR_ABZP']*1.
        fmin = 10**(-0.4*(33-ZP))
        fmax = 10**(-0.4*(12-ZP))
        #print('flux limits', fmin, fmax)
        
        for i, f in enumerate(photoz.flux_columns):
            key = f'mag_{f}'
            df[key] = ZP - 2.5*np.log10(np.clip(fill_masked(photoz.cat[f]*self.zp[i]), 
                                                fmin, fmax))
            # if hasattr(photoz.cat[f], 'mask'):
            #     
            #     df[key] = ZP - 2.5*np.log10(np.clip(photoz.cat[f].filled(-99), 
            #                                         fmin, fmax))
            # else:
            #     df[key] = ZP - 2.5*np.log10(np.clip(photoz.cat[f], 
            #                                         fmin, fmax))
        
        df['mag'] = df[f'mag_{self.DEFAULT_FILTER}']
        df['mag1mag2'] = df['mag']*0
        
        self.extra_plots = {}
        
        for k in extra_plots:
            _plot_args = extra_plots[k]
            xcol, ycol = _plot_args[:2]
            
            if xcol not in df.columns:
                df[xcol] = fill_masked(zout[xcol])
                    
            if ycol not in df.columns:
                df[ycol] = fill_masked(zout[ycol])

            if len(_plot_args) == 2:
                self.extra_plots[k] = (xcol, ycol, xcol, ycol, 
                                       np.nanmin(xcol), np.nanmax(ycol))
            elif len(_plot_args) == 4:
                self.extra_plots[k] = (*_plot_args, 
                                       np.nanmin(xcol), np.nanmax(ycol))
            elif len(_plot_args) == 6:
                self.extra_plots[k] = _plot_args
            else:
                print(f'Expected 2,4 or 6 elements in extra_plots[{k}], '
                      f'found {len(_plot_args)}')
        
        if selection is not None:
            self.df = df[selection].to_pandas()
        else:
            self.df = df.to_pandas()

        self.zout = zout
        self.photoz = photoz
        
        self.ZMAX = photoz.zgrid.max()
        self.MAXNFILT = photoz.nusefilt.max()
        

    @property
    def ra_bounds(self):
        return (self.df['ra'].max(), self.df['ra'].min())


    @property 
    def dec_bounds(self):
        return (self.df['dec'].min(), self.df['dec'].max())

    
    def get_filters_from_api(self):
        """
        Query available HST/JWST filters from the heroku API
        """
        import urllib
        import urllib.request
        import json 
        import PIL.Image
        
        _ra = self.ra_bounds
        _dec = self.dec_bounds
        
        cosd = np.cos(np.mean(_dec)/180*np.pi)
        dx = np.diff(_ra)[0]*cosd
        dy = np.diff(_dec)[0]
        si = np.maximum(dx, dy)*3600
        
        furl = 'https://grizli-cutout.herokuapp.com/overlap?'
        furl += f'ra={np.mean(_ra)}&dec={np.mean(_dec)}&size={si}'
        with open('/tmp/thumb.log','a') as fp:
            fp.write(furl+'\n')
        
        with urllib.request.urlopen(furl) as url:
            olap = json.loads(url.read().decode())
        
        return olap
    
    
    def make_dash_app(self, template='plotly_white', server_mode='external', port=8050, app=None, app_type='jupyter', plot_height=680, external_stylesheets=['https://codepen.io/chriddyp/pen/bWLwgP.css'], infer_proxy=False, slider_width=140, cutout_hdu=None, cutout_rgb=None, cutout_size=10, api_filters=None, api_size=2, api_args='', PLOT_TYPES=['zphot-zspec', 'Mag-redshift', 'Mag-color', 'redshift-color', 'Mass-redshift', 'UVJ', 'RA/Dec', 'UV-redshift', 'chi2-redshift'], get_content=False, fitsmap_url=None, cutout_url=CUTOUT_URL):
        """
        Create a Plotly/Dash app for interactive exploration
        
        Parameters
        ----------
        template : str
            `plotly` style `template <https://plotly.com/python/templates/#specifying-themes-in-graph-object-figures>`_.
        
        server_mode, port : str, int
            If not `None`, the app server is started with 
            `app.run_server(mode=server_mode, port=port)`.
        
        app_type : str
            If ``jupyter`` then `app = jupyter_dash.JupyterDash()`, else
            `app = dash.Dash()`
            
        plot_height : int
            Height in pixels of the scatter and SED+P(z) plot windows.
        
        infer_proxy : bool
            Run `JupyterDash.infer_jupyter_proxy_config()`, before app 
            initilization, e.g., for running on GoogleColab.
            
        Returns
        -------
        app : object
            App object following `app_type`.
            
        """
        import dash
        from dash import  dcc, html, callback
        from dash.dependencies import Input, Output
        import plotly.express as px
        
        import matplotlib.pyplot as plt
        from urllib.parse import urlparse, parse_qsl, urlencode
        import astropy.wcs as pywcs
        
        self.template = template
        self.plot_height = plot_height
        self.slider_width = slider_width
        self.cutout_size = cutout_size
        
        pivots = {}
        for f, w in zip(self.photoz.flux_columns, self.photoz.pivot):
            pivots[f] = w
        
        if app is None:
            if app_type == 'dash':
                app = dash.Dash(__name__, 
                                external_stylesheets=external_stylesheets)
            else:
                from jupyter_dash import JupyterDash
                if infer_proxy:
                    JupyterDash.infer_jupyter_proxy_config()
                
                app = JupyterDash(__name__, 
                                  external_stylesheets=external_stylesheets)
        
        for _t in self.extra_plots:
            PLOT_TYPES.append(_t)
            
        COLOR_TYPES = ['z_phot', 'z_spec', 'mass', 'sSFR', 'chi2']
        self.COLOR_TYPES = COLOR_TYPES
        self.PLOT_TYPES = PLOT_TYPES
        self.cutout_data = None
        self.cutout_wcs = None
        self.api_filters = api_filters
        self.fitsmap_url = fitsmap_url
        self.cutout_url = cutout_url
        
        #_title = f"{self.photoz.param['MAIN_OUTPUT_FILE']}"
        #_subhead = f"Nobj={self.photoz.NOBJ}  Nfilt={self.photoz.NFILT}"
        _title = [html.Strong(self.photoz.param['MAIN_OUTPUT_FILE']), 
                  ' / N', html.Sub('obj'), f'={self.photoz.NOBJ}', 
                  ' / N', html.Sub('filt'), f'={self.photoz.NFILT}', 
                  ]
        
        slider_row_style={'width': '90%', 'float':'left', 
                          'margin-left':'10px'}          
        slider_container = {'width': f'{slider_width}px', 
                            'margin-left':'-25px'}
        check_kwargs = dict(style={'text-align':'center', 
                                   'height':'14pt', 
                                   'margin-top':'-20px'})
        
        # bool_options = {'has_zspec': 'z_spec > 0', 
        #                 'use': 'Use == 1'}
        
        if cutout_hdu is not None:
            self.cutout_wcs = pywcs.WCS(cutout_hdu.header, relax=True)
            if cutout_rgb is None:
                self.cutout_data = cutout_hdu.data
            else:
                self.cutout_data = np.flipud(plt.imread(cutout_rgb))
                
            print('xxx', self.cutout_data.shape)
            
            cutout_div = html.Div([
                             dcc.Graph(id='cutout-figure', 
                                       style={})
                                 ], style={'right':'70px', 
                                           'width':'120px',
                                           'height':'120px',
                                        'border':'1px solid rgb(200,200,200)',
                                           'top':'10px', 
                                           'position':'absolute'})
            cutout_target = 'figure'
            
        elif api_filters is not None:
            
            cutout_div = html.Div([
                             dcc.Graph(id='cutout-figure', 
                                       style={})
                                 ], style={'right':'70px', 
                                           'width':'120px',
                                           'height':'120px',
                                        'border':'1px solid rgb(200,200,200)',
                                           'top':'10px', 
                                           'position':'absolute'})
            cutout_target = 'figure'
            self.cutout_data = None
            
        else:
            cutout_div = html.Div(id='cutout-figure', 
                                   style={'left':'1px', 
                                           'width':'1px',
                                           'height':'1px',
                                           'bottom':'1px', 
                                           'position':'absolute'})
            self.cutout_data = None
            cutout_target = 'children'
                
        
        ####### Main content
        content = [
            # Selectors
            html.Div([
                dcc.Location(id='url', refresh=False), 
                
                html.Div([
                    html.Div(_title, id='title-bar', 
                             style={'float':'left', 'margin-top':'4pt'}),
                    
                    html.Div([
                        html.Div([dcc.Dropdown(id='plot-type',
                                     options=[{'label': i, 'value': i}
                                              for i in PLOT_TYPES],
                                     value=PLOT_TYPES[0], 
                                     clearable=False,
                                     style={'width':'120px', 
                                            'margin-right':'5px',
                                            'margin-left':'5px',
                                            'font-size':'8pt'}),
                        ], style={'float':'left'}),
                        
                        html.Div([dcc.Dropdown(id='color-type',
                                     options=[{'label': i, 'value': i}
                                              for i in COLOR_TYPES],
                                     value='sSFR', 
                                     clearable=False,
                                     style={'width':'80px', 
                                            'margin-right':'5px',
                                            'font-size':'8pt'}),
                        ], style={'display':'inline-block', 
                                  'margin-left':'10px'}),
                    ], style={'float':'right'}),
                ], style=slider_row_style),
                
                html.Div([
                    html.Div([dcc.Dropdown(id='mag-filter',
                                 options=[{'label': i, 'value': i}
                                          for i in self.photoz.flux_columns],
                                 value=self.DEFAULT_FILTER, 
                                 style={'width': f'{slider_width-45}px', 
                                        'margin-right':'20px',
                                        'font-size':'8pt'},
                                 clearable=False),
                    ], style={'float':'left'}),
                                        
                    html.Div([
                        dcc.RangeSlider(id='mag-slider',
                                        min=12, max=32, step=0.2,
                                        value=[18, 27],
                                        updatemode='drag',
                                        tooltip={"placement":'left'}, 
                                        marks=None), 

                        dcc.Checklist(id='mag-checked', 
                                      options=[{'label':'AB mag', 
                                                'value':'checked'}], 
                                      value=['checked'], **check_kwargs),
                              
                    ], style=dict(display='inline-block',
                                  **slider_container)),
                    
                    # Mag2 for color
                    html.Div([dcc.Dropdown(id='mag2-filter',
                                 options=[{'label': i, 'value': i}
                                          for i in self.photoz.flux_columns],
                                 value=self.photoz.flux_columns[0], 
                                 style={'width': f'{slider_width-45}px', 
                                        'margin-right':'20px',
                                        'font-size':'8pt'},
                                 clearable=False),
                    ], style={'float':'left'}),
                    
                    html.Div([
                        dcc.RangeSlider(id='color-slider',
                                        min=-2, max=5, step=0.1,
                                        value=[-1, 3],
                                        updatemode='drag',
                                        tooltip={"placement":'left'}, 
                                        marks=None), 

                        dcc.Checklist(id='color-checked', 
                                      options=[{'label':'Mag1 - Mag2', 
                                                'value':'checked'}], 
                                      value=[], **check_kwargs),
                              
                    ], style=dict(display='inline-block',
                                  **slider_container)),
                    
                    html.Div([
                        dcc.RangeSlider(id='nfilt-slider',
                                        min=1, max=self.MAXNFILT, step=1,
                                        value=[3, self.MAXNFILT],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),

                        dcc.Checklist(id='nfilt-checked', 
                                      options=[{'label':'nfilt', 
                                                'value':'checked'}], 
                                      value=['checked'], **check_kwargs),

                    ], style=dict(display='inline-block', 
                                  **slider_container)),
                    
                ], style=slider_row_style),
                
                html.Div([
                    html.Div([
                        dcc.RangeSlider(id='zphot-slider',
                                        min=-0.5, max=self.ZMAX, step=0.1,
                                        value=[0, self.ZMAX],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),
                          
                         dcc.Checklist(id='zphot-checked', 
                                       options=[{'label':'z_phot', 
                                                 'value':'checked'}], 
                                       value=['checked'], **check_kwargs),
                
                    ], style=dict(float='left', **slider_container)), 

                    html.Div([
                        dcc.RangeSlider(id='zspec-slider',
                                        min=-0.5, max=self.ZMAX, step=0.1,
                                        value=[-0.5, 6.5],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),
                          
                        dcc.Checklist(id='zspec-checked', 
                                      options=[{'label':'z_spec', 
                                                'value':'checked'}], 
                                      value=['checked'], **check_kwargs),

                    ], style=dict(display='inline-block',
                                  **slider_container)),

                    html.Div([
                        dcc.RangeSlider(id='mass-slider',
                                        min=7, max=13, step=0.1,
                                        value=[8, 11.8],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),

                        dcc.Checklist(id='mass-checked', 
                                      options=[{'label':'mass', 
                                                'value':'checked'}], 
                                      value=['checked'], **check_kwargs),

                    ], style=dict(display='inline-block', 
                                  **slider_container)),
                    
                    html.Div([
                        dcc.RangeSlider(id='chi2-slider',
                                        min=0, max=20, step=0.1,
                                        value=[0, 6],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),

                        dcc.Checklist(id='chi2-checked', 
                                      options=[{'label':'chi2',
                                                'value':'checked'}], 
                                      value=[], **check_kwargs),

                    ], style=dict(display='inline-block', 
                                  **slider_container)),
                    
                    # Customizable slider
                    html.Div([
                        dcc.RangeSlider(id='extra-slider',
                                        min=self.extra_slider_kws['min'], 
                                        max=self.extra_slider_kws['max'],
                                        step=self.extra_slider_kws['step'],
                                        value=self.extra_slider_kws['value'],
                                        updatemode='drag',
                                        tooltip={"placement":'left'},
                                        marks=None),

                        dcc.Checklist(id='extra-checked', 
                                      options=[{'label':self.extra_slider_kws['label'], 
                                                'value':'checked'}], 
                                      value=self.extra_slider_kws['checked'], 
                                      **check_kwargs),

                    ], style=dict(display='inline-block', 
                                  **slider_container)),
                    
                    # Boolean dropdown
                    # dcc.Dropdown(id='bool-checks',
                    #              options=[{'label': self.bool_options[k], 
                    #                        'value': k}
                    #                       for k in self.bool_options],
                    #              value=[], 
                    #              multi=True,
                    #              style={'width':'100px', 
                    #                     'display':'inline-block',
                    #                     'margin-left':'0px', 
                    #                     'font-size':'8pt'},
                    #              clearable=True),
                    
                ], style=slider_row_style), 

            ], style={'float':'left','width': '55%'}),

            # Object-level controls
            html.Div([
                html.Div([
                    html.Div('ID / RA,Dec.', style={'float':'left', 
                                                       'width':'100px', 
                                                       'margin-top':'5pt'}), 
                    
                    dcc.Input(id='id-input', type='text',
                          style={'width':'120px',
                                 'padding':'2px', 
                                 'display':'inline',
                                 'font-size':'8pt'}), 
                    html.Div(children='', id='match-sep', 
                           style={'margin':'5pt',
                           'display':'inline', 
                                  'width':'50px',
                                  'font-size':'8pt'}),
                    
                    dcc.RadioItems(id='sed-unit-selector',
                                             options=[{'label': i, 'value': i}
                                             for i in ['Fλ', 'Fν', 'νFν']],
                                             value='Fλ',
                                             labelStyle={'display':'inline', 
                                                         'padding':'3px', 
                                                         },
                                             style={'display':'inline',
                                                    'width':'130px'})
                                                              
                ],  style={'width':'260pix', 'float':'left', 
                           'margin-right':'20px'}),
                
            ]), 
            html.Div([           
                # html.Div([
                # ],  style={'width':'120px', 'float':'left'}), 

                html.Div(id='object-info', children='ID: ', 
                        style={'margin':'auto','margin-top':'10px', 
                               'font-size':'10pt'})

            ], style={'float':'right', 'width': '45%'}),

            # Plots
            html.Div([# Scatter plot
                dcc.Graph(id='sample-selection-scatter', 
                          hoverData={'points': [{'customdata':
                                             (self.df['id'][0], 
                                              1.0, -9.0)}]}, 
                          style={'width':'95%'})
            ], style={'float':'left', 'height':'70%', 'width':'49%'}), 

            html.Div([# SED
                dcc.Graph(id='object-sed-figure',
                          style={'width':'95%'})
            ], style={'float':'right', 'width':'49%', 'height':'70%'}),
            
            cutout_div
        ]

        if get_content:
            return content
        
        ##### Callback functions
        @app.callback(
             Output('url', 'search'),
            [Input('plot-type', 'value'),
             Input('color-type', 'value'),
             Input('mag-filter', 'value'),
             Input('mag-slider', 'value'),
             Input('mag2-filter', 'value'),
             Input('color-slider', 'value'),
             Input('mass-slider', 'value'),
             Input('chi2-slider', 'value'),
             Input('nfilt-slider', 'value'),
             Input('zphot-slider', 'value'),
             Input('zspec-slider', 'value'),
             Input('id-input', 'value')])
        def update_url_state(plot_type, color_type, mag_filter, mag_range, mag2_filter, color_range, mass_range, chi2_range, nfilt_range, zphot_range, zspec_range, id_input):
            search = f'?plot_type={plot_type}&color_type={color_type}'
            search += f'&mag_filter={mag_filter}'
            search += f'&mag={mag_range[0]},{mag_range[1]}'
            search += f'&mag2_filter={mag2_filter}'
            search += f'&color={color_range[0]},{color_range[1]}'
            search += f'&mass={mass_range[0]},{mass_range[1]}'
            search += f'&chi2={chi2_range[0]},{chi2_range[1]}'
            search += f'&nfilt={nfilt_range[0]},{nfilt_range[1]}'
            search += f'&zphot={zphot_range[0]},{zphot_range[1]}'
            search += f'&zspec={zspec_range[0]},{zspec_range[1]}'
            if id_input is not None:
                search += f"&id={id_input.replace(' ', '%20')}"
                
            return search


        @app.callback([Output('plot-type', 'value'),
                       Output('color-type', 'value'),
                       Output('mag-filter', 'value'),
                       Output('mag-slider', 'value'),
                       Output('mag2-filter', 'value'),
                       Output('color-slider', 'value'),
                       Output('mass-slider', 'value'),
                       Output('chi2-slider', 'value'),
                       Output('nfilt-slider', 'value'),
                       Output('zphot-slider', 
                                                'value'),
                       Output('zspec-slider', 
                                                'value'),
                       Output('id-input', 'value'),
                      ],[
                       Input('url', 'search')
                      ])
        def set_state_from_url(search):
            plot_type = PLOT_TYPES[0]
            color_type = 'sSFR'
            mag_filter = self.DEFAULT_FILTER
            mag_range = [18, 27]
            mag2_filter = self.DEFAULT_FILTER
            color_range = [-0.5, 3]
            mass_range = [8, 11.6]
            chi2_range = [0, 4]
            nfilt_range = [1, self.MAXNFILT]
            zphot_range = [0, self.ZMAX]
            zspec_range = [-0.5, 6.5]
            id_input = None

            # if '?' not in href:
            if not search:
                return (plot_type, color_type, mag_filter, mag_range,
                        mag2_filter, color_range,
                        mass_range, chi2_range, nfilt_range,
                        zphot_range, zspec_range,
                        id_input)

            # search = href.split('?')[1]
            params = search.split('&')

            for p in params:
                if 'plot_type' in p:
                    val = p.split('=')[1]
                    if val in PLOT_TYPES:
                        plot_type = val

                elif 'color_type' in p:
                    val = p.split('=')[1]
                    if val in COLOR_TYPES:
                        color_type = val
                        
                elif 'mag_filter' in p:
                    val = p.split('=')[1]
                    if val in self.photoz.flux_columns:
                        mag_filter = val
                    
                elif 'mag=' in p:
                    try:
                        vals = [float(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            mag_range = vals
                    except ValueError:
                        pass
                        
                elif 'mag2_filter' in p:
                    val = p.split('=')[1]
                    if val in self.photoz.flux_columns:
                        mag2_filter = val

                elif 'color=' in p:
                    try:
                        vals = [float(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            color_range = vals
                    except ValueError:
                        pass

                elif 'mass' in p:
                    try:
                        vals = [float(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            mass_range = vals
                    except ValueError:
                        pass
                
                elif 'nfilt=' in p:
                    try:
                        vals = [int(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            nfilt_range = vals
                    except ValueError:
                        pass
                
                elif 'zspec' in p:
                    try:
                        vals = [float(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            zspec_range = vals
                    except ValueError:
                        pass
                        
                elif 'zphot' in p:
                    try:
                        vals = [float(v) for v in p.split('=')[1].split(',')]
                        if len(vals) == 2:
                            zphot_range = vals
                    except ValueError:
                        pass
                        
                elif 'id' in p:
                    try:
                        id_input = p.split('=')[1].replace('%20', ' ')
                    except ValueError:
                        id_input = None
                    
                    if not id_input:
                        id_input = None
                        
            return (plot_type, color_type, mag_filter, mag_range,
                    mag2_filter, color_range,
                    mass_range, chi2_range, nfilt_range,
                    zphot_range, zspec_range,
                    id_input)

        @app.callback(
            Output('nfilt-checked', 'label'),
            Input('sample-selection-scatter', 'selectedData'))
        def update_selected_data(selectedData):
            
            if selectedData is not None:
                #print('selectedData', len(selectedData['points']))
                
                #print('y0', self.df['in_selectedData'].sum())
                
                if len(selectedData['points']) == 1:
                    # Reset
                    self.df['in_selectedData'] = self.df['z_phot'] > 0
                    print('One point selected: reset selection!')
                    
                elif len(selectedData['points']) > 0:
                    selected_ids = []
                    for p in selectedData['points']:
                        if "customdata" in p.keys():
                            selected_ids.append(p['customdata'][0])
                    
                    #print(len(selected_ids), p)
                    
                    if len(selected_ids) == 1:
                        # Reset
                        self.df['in_selectedData'] = self.df['z_phot'] > 0
                        print('One point selected: reset selection!')
                        
                    elif len(selected_ids) > 0:
                        self.df['in_selectedData'] = np.isin(self.df['id'], 
                                                             selected_ids)
                
                #print('y1', self.df['in_selectedData'].sum())
                
            else:
                # print('selectedData None')
                self.df['in_selectedData'] = self.df['z_phot'] > 0
                
            return 'nfilt' #f"N: {self.df['in_selectedData'].sum()}"
        
        
        @app.callback(
            Output('sample-selection-scatter', 'figure'),
            [Input('plot-type', 'value'),
             Input('color-type', 'value'), 
             Input('mag-filter', 'value'),
             Input('mag-slider', 'value'),
             Input('mag-checked', 'value'),
             Input('mag2-filter', 'value'),
             Input('color-slider', 'value'),
             Input('color-checked', 'value'),
             Input('mass-slider', 'value'),
             Input('mass-checked', 'value'),
             Input('extra-slider', 'value'),
             Input('extra-checked', 'value'),
             Input('chi2-slider', 'value'),
             Input('chi2-checked', 'value'),
             Input('nfilt-slider', 'value'),
             Input('nfilt-checked', 'value'),
             Input('zphot-slider', 'value'),
             Input('zphot-checked', 'value'),
             Input('zspec-slider', 'value'),
             Input('zspec-checked', 'value'),
             Input('id-input', 'value')
         ])
        def update_selection(plot_type, color_type, mag_filter, mag_range, mag_checked, mag2_filter, color_range, color_checked, mass_range, mass_checked, extra_range, extra_checked, chi2_range, chi2_checked, nfilt_range, nfilt_checked, zphot_range, zphot_checked, zspec_range, zspec_checked, id_input):
            """
            Apply slider selections
            """
            sel = np.isfinite(self.df['z_phot'])
            if 'checked' in zphot_checked:
                sel &= (self.df['z_phot'] > zphot_range[0]) 
                sel &= (self.df['z_phot'] < zphot_range[1])
            
            if 'checked' in zspec_checked:
                sel &= (self.df['z_spec'] > zspec_range[0]) 
                sel &= (self.df['z_spec'] < zspec_range[1])
            
            if 'checked' in mass_checked:
                sel &= (self.df['mass'] > mass_range[0])
                sel &= (self.df['mass'] < mass_range[1])

            if 'checked' in chi2_checked:
                sel &= (self.df['chi2'] >= chi2_range[0])
                sel &= (self.df['chi2'] <= chi2_range[1])

            if 'checked' in nfilt_checked:
                sel &= (self.df['nusefilt'] >= nfilt_range[0])
                sel &= (self.df['nusefilt'] <= nfilt_range[1])

            if 'checked' in extra_checked:
                _col = self.extra_slider_kws['column']
                sel &= (self.df[_col] >= extra_range[0])
                sel &= (self.df[_col] <= extra_range[1])
            
            #print('redshift: ', sel.sum())
            
            if mag_filter is None:
                mag_filter = self.DEFAULT_FILTER
            
            if mag2_filter is None:
                mag2_filter = self.DEFAULT_FILTER
            
            #self.self.df['mag'] = self.ABZP 
            #self.self.df['mag'] -= 2.5*np.log10(self.photoz.cat[mag_filter])
            mag_col = 'mag_'+mag_filter            
            if 'checked' in mag_checked:
                sel &= (self.df[mag_col] > mag_range[0]) 
                sel &= (self.df[mag_col] < mag_range[1])
            
            mag2_col = 'mag_'+mag2_filter
            
            fblue, fred = get_sorted_mag_columns(mag_filter, mag2_filter)
            
            mag1mag2 = self.df['mag_'+fblue] - self.df['mag_'+fred]
            
            if 'checked' in color_checked:
                sel &= (mag1mag2 > color_range[0]) 
                sel &= (mag1mag2 < color_range[1])
            
            self.df['mag'] = self.df[mag_col]
            self.df['mag1mag2'] = mag1mag2
            
            #print('mag: ', sel.sum())
            
            if plot_type == 'zphot-zspec':
                sel &= self.df['z_spec'] > 0
            
            #print('zspec: ', sel.sum())
            
            if id_input is not None:
                id_i, dr_i = parse_id_input(id_input)
                if id_i is not None:
                    self.df['is_selected'] = self.df['id'] == id_i
                    sel |= self.df['is_selected']
                else:
                    self.df['is_selected'] = False
            else:
                self.df['is_selected'] = False
            
            xsel = self.df['in_selectedData'][sel]
            dff = self.df[sel]
            
            # Color-coding by color-type pulldown
            if color_type == 'z_phot':
                color_kwargs = dict(color=np.clip(dff['z_phot'][xsel], 
                                                  *zphot_range),
                                    color_continuous_scale='portland')
            elif color_type == 'z_spec':
                color_kwargs = dict(color=np.clip(dff['z_spec'][xsel], 
                                                  *zspec_range), 
                                    color_continuous_scale='portland')
            elif color_type == 'mass':
                color_kwargs = dict(color=np.clip(dff['mass'][xsel], *mass_range), 
                                    color_continuous_scale='magma_r')
            elif color_type == 'chi2':
                color_kwargs = dict(color=np.clip(dff['chi2'][xsel], *chi2_range), 
                                    color_continuous_scale='viridis')            
            else:
                color_kwargs = dict(color=np.clip(dff['ssfr'][xsel], -12., -8.), 
                                    color_continuous_scale='portland_r')
            
            # Scatter plot  
            plot_defs = {'Mass-redshift':['z_phot','mass',
                                     'z<sub>phot</sub>', 'log Stellar mass', 
                                     (-0.1, self.ZMAX), (7.5, 12.5)], 
                         'Mag-redshift': ['z_phot','mag',
                                'z<sub>phot</sub>', f'AB mag ({mag_filter})', 
                                (-0.1, self.ZMAX), (18, 28)],
                         'Mag-color': ['mag','mag1mag2',
                                f'AB mag ({mag_filter})',
                                f'{fblue} - {fred}'.replace('_tot_1',''),
                                (18, 28),
                                (-0.5, 3)],
                         'redshift-color': ['z_phot','mag1mag2',
                                f'z_phot',
                                f'{fblue} - {fred}'.replace('_tot_1',''),
                                (-0.1, self.ZMAX),
                                (-0.5, 3)],
                         'RA/Dec': ['ra','dec',
                                    'R.A.', 'Dec.', 
                                    self.ra_bounds, self.dec_bounds], 
                         'zphot-zspec': ['z_spec','z_phot',
                                   'z<sub>spec</sub>', 'z<sub>phot</sub>', 
                                    (0, 4.5), (0, 4.5)], 
                         'UVJ': ['vj','uv',
                                  '(V-J)', '(U-V)', 
                                  (-0.1, 2.5), (-0.1, 2.5)], 
                         'UV-redshift': ['z_phot','uv',
                                 'z<sub>phot</sub>', '(U-V)<sub>rest</sub>', 
                                 (0, 4), (-0.1, 2.50)], 
                         'chi2-redshift': ['z_phot','chi2',
                                 'z<sub>phot</sub>', 'chi<sup>2</sup>',
                                 (0, 4), (0.1, 30)]
                         }
            
            if plot_type in self.extra_plots:
                args = [*self.extra_plots[plot_type], {}, color_kwargs]
            elif plot_type in plot_defs:
                args = [*plot_defs[plot_type], {}, color_kwargs]
            else:
                args = [*plot_defs['zphot-zspec'], {}, color_kwargs]
            
            if args[0] == 'mag':
                args[2] = f'AB mag ({mag_filter})'
            
            if args[1] == 'mag':
                args[3] = f'AB mag ({mag_filter})'
                
            fig = update_sample_scatter(dff, *args)
            
            # Update ranges for some parameters
            if ('Mass' in plot_type) & ('checked' in mass_checked):
                fig.update_yaxes(range=mass_range)

            if ('Mag' in plot_type) & ('checked' in mag_checked):
                if args[0] == 'mag':
                    fig.update_xaxes(range=mag_range)
                else:
                    fig.update_yaxes(range=mag_range)
                    
            if ('color' in plot_type) & ('checked' in color_checked):
                if args[0] == 'mag1mag2':
                    fig.update_xaxes(range=color_range)
                else:
                    fig.update_yaxes(range=color_range)
                    
            if ('redshift' in plot_type) & ('checked' in zphot_checked):
                if args[0] == 'z_phot':
                    fig.update_xaxes(range=zphot_range)
                else:
                    fig.update_yaxes(range=zphot_range)
                    
            if ('zspec' in plot_type) & ('checked' in zspec_checked):
                if args[0] == 'z_spec':
                    fig.update_xaxes(range=zspec_range)
                else:
                    fig.update_yaxes(range=zspec_range)
            
            return fig


        def update_sample_scatter(dff, xcol, ycol, x_label, y_label, x_range, y_range, extra,  color_kwargs):
            """
            Make scatter plot
            """
            import plotly.graph_objects as go
            
            # print('update_sample_scatter xxx', xcol, len(dff[xcol]))
            is_sel = dff['in_selectedData']
            fig = px.scatter(data_frame=dff[is_sel], x=xcol, y=ycol, 
                             custom_data=['id','z_phot','mass','ssfr','mag'], 
                             **color_kwargs)
                            
            htempl = '(%{x:.2f}, %{y:.2f}) <br>'
            htempl += 'id: %{customdata[0]:0d}  z_phot: %{customdata[1]:.2f}'
            htempl += '<br> mag: %{customdata[4]:.1f}  '
            htempl += 'mass: %{customdata[2]:.2f}  ssfr: %{customdata[3]:.2f}'
                
            fig.update_traces(hovertemplate=htempl, opacity=0.7)

            if (~is_sel).sum() > 0:
                _xsel = go.Scatter(x=dff[~is_sel][xcol],
                                   y=dff[~is_sel][ycol], 
                                   mode="markers", 
                                   marker=dict(color='rgba(180,180,180,0.2)',
                                               size=5, symbol='circle'),
                                   hoverinfo='skip',
                               )
                fig.add_trace(_xsel)
                fig.data = fig.data[::-1]

            if dff['is_selected'].sum() > 0:
                dffs = dff[dff['is_selected']]
                _sel = go.Scatter(x=dffs[xcol], y=dffs[ycol],
                                  mode="markers+text",
                                  text=[f'{id}' for id in dffs['id']],
                                  textposition="bottom center",
                                  marker=dict(color='rgba(250,0,0,0.5)', 
                                              size=20, 
                                              symbol='circle-open'))
                                  
                fig.add_trace(_sel)
            
            
            fig.update_xaxes(range=x_range, title_text=x_label)
            fig.update_yaxes(range=y_range, title_text=y_label)

            fig.update_layout(template=template, 
                              autosize=True, showlegend=False, 
                              margin=dict(l=0,r=0,b=0,t=20,pad=0,
                                          autoexpand=True))
            
            if plot_height is not None:
                fig.update_layout(height=plot_height)
                
            fig.update_traces(marker_showscale=False, 
                              selector=dict(type='scatter'))
            fig.update_coloraxes(showscale=False)
            
            if (xcol, ycol) == ('z_spec','z_phot'):
                _one2one = go.Scatter(x=[0, 8], y=[0,8],
                                  mode="lines",
                                  marker=dict(color='rgba(250,0,0,0.5)'))
                fig.add_trace(_one2one)
                
            fig.add_annotation(text=f'N = {len(dff)} / {len(self.df)}',
                          xref="x domain", yref="y domain",
                          x=0.98, y=0.05, showarrow=False)
            
            return fig


        def heroku_thumbnail(id_i):
            """
            Thumbnail from grizli API
            """
            import urllib
            import urllib.request
            import json 
            import PIL.Image
            
            ix = np.where(self.df['id'] == id_i)[0][0]
            ri, di = self.df['ra'][ix], self.df['dec'][ix]
            
            turl = f'https://grizli-cutout.herokuapp.com/thumb?'
            turl += f'ra={ri}&dec={di}&size={api_size}'
            turl += f'{api_args}&filters={api_filters}'   
            print(turl)
            # with open('/tmp/thumb.log','a') as fp:
            #     fp.write(turl+'\n')
                
            req = urllib.request.urlopen(turl)
                                                        
            thumb = np.array(PIL.Image.open(req))
            
            return thumb

        
        def api_cutout_figure(id_i):
            """
            Thumbnail from grizli API
            """
            try:
                cutout = heroku_thumbnail(id_i)
            except:
                cutout = np.zeros((10, 10))
            
            sh = cutout.shape
            
            fig = px.imshow(cutout, origin='upper')

            fig.update_coloraxes(showscale=False)
            fig.update_layout(width=120, height=120, 
                              margin=dict(l=0,r=0,b=0,t=0,pad=0,
                                              autoexpand=True))

            fig.update_xaxes(range=(-0.5, sh[1]-0.5), 
                             visible=False, showticklabels=False)
            fig.update_yaxes(range=(-0.5, sh[0]-0.5),
                             visible=False, showticklabels=False)
            
            return fig


        def hdu_cutout_figure(id_i):
            """
            SED cutout
            """

            ix = np.where(self.df['id'] == id_i)[0]
            ri, di = self.df['ra'][ix], self.df['dec'][ix]
            xi, yi = np.squeeze(self.cutout_wcs.all_world2pix([ri], [di], 0))
            xp = int(np.round(xi))
            yp = int(np.round(yi))
            slx = slice(xp-cutout_size,xp+cutout_size+1)
            sly = slice(yp-cutout_size,yp+cutout_size+1)

            try:
                if self.cutout_data.ndim == 2:
                    cutout = self.cutout_data[sly, slx]
                    fig = px.imshow(cutout, color_continuous_scale='gray_r', 
                                origin='lower')
                else:
                    cutout = self.cutout_data[sly, slx, :]
                    fig = px.imshow(cutout, origin='lower')
            except:
                cutout = np.zeros((2*cutout_size, 2*cutout_size))
                
                fig = px.imshow(cutout, color_continuous_scale='gray_r', 
                            origin='lower')

            fig.update_coloraxes(showscale=False)
            fig.update_layout(width=120, height=120, 
                              margin=dict(l=0,r=0,b=0,t=0,pad=0,
                                              autoexpand=True))

            fig.update_xaxes(range=(0, 2*cutout_size), 
                             visible=False, showticklabels=False)
            fig.update_yaxes(range=(0, 2*cutout_size),
                             visible=False, showticklabels=False)

            return fig

        
        def get_sorted_mag_columns(mag_filter, mag2_filter):
            """
            """
            if pivots[mag_filter] > pivots[mag2_filter]:
                return mag2_filter, mag_filter
            else:
                return mag_filter, mag2_filter
        
        def get_map_link(ra, dec):
            """
            Return an html.A object with a link to a FITSMap or LegacySurvey map
            """
            from dash import html
            
            if self.fitsmap_url is None:
                link = html.A('LegacySurvey', 
                              href=utils.show_legacysurvey(ra, dec, layer='ls-dr9'))
            else:
                href = self.fitsmap_url.format(ra=ra, dec=dec)
                link = html.A('FitsMap', href=href)
            
            return link
        
        
        def parse_id_input(id_input):
            """
            Parse input as id or (ra dec)
            """
            if id_input in ['None', None, '']:
                return None, None
            
            inp_split = id_input.replace(',',' ').split()
            
            if len(inp_split) == 1:
                return int(inp_split[0]), None
                
            ra, dec = np.asarray(inp_split, dtype=float)
            
            cosd = np.cos(self.df['dec']/180*np.pi)
            dx = (self.df['ra'] - ra)*cosd
            dy = (self.df['dec'] - dec)
            dr = np.sqrt(dx**2+dy**2)*3600.
            imin = np.nanargmin(dr)
            
            return self.df['id'][imin], dr[imin]
            

        @app.callback([Output('object-sed-figure', 'figure'),
                       Output('object-info', 'children'), 
                       Output('match-sep', 'children'), 
                       Output('cutout-figure', cutout_target)], 
                      [Input('sample-selection-scatter', 
                                               'hoverData'), 
                       Input('sed-unit-selector', 'value'),
                       Input('id-input', 'value')])
        def update_object_sed(hoverData, sed_unit, id_input):
            """
            SED + p(z) plot
            """
            id_i, dr_i = parse_id_input(id_input)
            if id_i is None:
                id_i = hoverData['points'][0]['customdata'][0]
            else:
                if id_i not in self.zout['id']:
                    id_i = hoverData['points'][0]['customdata'][0]
            
            if dr_i is None:
                match_sep = ''
            else:
                match_sep = f'{dr_i:.1f}"'
                        
            show_fnu = {'Fλ':0, 'Fν':1, 'νFν':2}
            
            layout_kwargs = dict(template=template, 
                                 autosize=True, 
                                 showlegend=False, 
                                 margin=dict(l=0,r=0,b=0,t=20,pad=0,
                                               autoexpand=True))
                              
            fig = self.photoz.show_fit_plotly(id_i,
                                              show_fnu=show_fnu[sed_unit], 
                                              vertical=True, 
                                              panel_ratio=[0.6, 0.4],
                                              show=False,
                                              layout_kwargs=layout_kwargs)
            
            if plot_height is not None:
                fig.update_layout(height=plot_height)
            
            
            ix = self.df['id'] == id_i
            if ix.sum() == 0:
                object_info = 'ID: N/A'
            else:
                ix = np.where(ix)[0][0]
                ra, dec = self.df['ra'][ix], self.df['dec'][ix]
                object_info = [f'ID: {id_i}  |  α, δ = {ra:.6f} {dec:.6f} ',
                               # ' | ', html.A('ESO',
                               #               href=utils.eso_query(ra, dec,
                               #                             radius=1.0,
                               #                             unit='s')),
                               ' | ', html.A('CDS', 
                                             href=utils.cds_query(ra, dec, 
                                                           radius=1.0,
                                                           unit='s')),
                               ' | ', html.A('Cutout',
                                          href=self.cutout_url.format(ra=ra, dec=dec)),
                               ' | ', get_map_link(ra, dec),
                               html.Br(), 
                               f"z_phot: {self.df['z_phot'][ix]:.3f}  ", 
                               f" | z_spec: {self.df['z_spec'][ix]:.3f}", 
                               html.Br(),  
                               f"mag: {self.df['mag'][ix]:.2f}  ", 
                               f" | mass: {self.df['mass'][ix]:.2f} ",
                               f" | sSFR: {self.df['ssfr'][ix]:.2f}", 
                               html.Br()]
            
            
            if self.cutout_data is not None:
                cutout_fig = hdu_cutout_figure(id_i)
            elif api_filters is not None:
                cutout_fig = api_cutout_figure(id_i)
            else:
                cutout_fig = ['']
                
            return fig, object_info, match_sep, cutout_fig
        
        ####### App layout
        app.layout = html.Div(content)
        
        if server_mode is not None:
            # app.run_server(mode=server_mode, port=port)
            app.run(jupyter_mode=server_mode, port=port)
        
        return app    


